/*####COPYRIGHTBEGIN####
 -------------------------------------------
 Copyright (C) 2005 Wapice Ltd

 Author: Kåre Särs (kare.sars@wapice.com)

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to:
 The Free Software Foundation, Inc.
 59 Temple Place - Suite 330
 Boston, MA  02111-1307
 USA.

 As a special exception, if other files instantiate templates or
 use macros or inline functions from this file, or you compile
 this file and link it with other works to produce a work based
 on this file, this file does not by itself cause the resulting
 work to be covered by the GNU General Public License. However
 the source code for this file must still be made available in
 accordance with section (3) of the GNU General Public License.

 This exception does not invalidate any other reasons why a work
 based on this file might be covered by the GNU General Public
 License.
 -------------------------------------------
####COPYRIGHTEND####*/

/* NOTE: This implementation does NOT support segmented messages
   and thus the aplicationlayer is not complete */

#include "platform.h"
#include "conf.h"
#include "npdu.h"
#include "apdu.h"
#include "bacenum.h"
#include "object.h"
#include "make_apdu.h"
#include "services.h"

/* APDU[0] */
#define apdu_is_segmented(apdu_0) ((apdu_0 & APDU_SEGMENTED) != 0)
#define apdu_has_more(apdu_0)     ((apdu_0 & APDU_MOR_SEGS) != 0)
#define apdu_segm_accept(apdu_0)  ((apdu_0 & APDU_SEG_ACC) != 0)

/* Confirmed Request */
/* Max Segments accepted */
#define APDU_MAX_SEGS_MASK   0x70
#define APDU_MAX_SEGS_UN_SP  0x00
#define APDU_MAX_SEGS_2      0x10
#define APDU_MAX_SEGS_4      0x20
#define APDU_MAX_SEGS_8      0x30
#define APDU_MAX_SEGS_16     0x40
#define APDU_MAX_SEGS_32     0x50
#define APDU_MAX_SEGS_64     0x60
#define APDU_MAX_SEGS_64PL   0x70

/* Max APDU Length */
#define APDU_MAX_LEN_MASK  0x0F
#define APDU_MAX_LEN_50    0x00
#define APDU_MAX_LEN_128   0x01
#define APDU_MAX_LEN_206   0x02
#define APDU_MAX_LEN_480   0x03
#define APDU_MAX_LEN_1024  0x04
#define APDU_MAX_LEN_1476  0x05

/* this is only a simplified state mashine */
typedef enum {
    APPL_LAYER_STATE_IDLE ,
    APPL_LAYER_STATE_SEGM_RESP,
    APPL_LAYER_STATE_SEGD_CONF
} APPL_LAYER_STATES;

struct appl_sruct_t {
    /* A variable to hold the Rx state */
    APPL_LAYER_STATES State;
    /* Used to perform timeout on PDU segments */
    INT16U            SegmentTimer;

    /* segmentation variables */
    INT16U            RequestedAPDUSize;
    INT8U             SequenceNumber;
    INT8U             InvokeId;
    INT8U             Data[MAX_APDU_LEN];
    INT16U            DataLen;
};

/* Windowed segmentation is not supported => propose ws=1, actual ws= 1 */
#define PROPOSED_WS 1
#define ACTUAL_WS   1

/****** localy global variables ********************************************/
static struct appl_sruct_t appl_rx_struct;
#ifdef SUPPORT_READ_PROPERTY_INIT
static struct appl_sruct_t appl_tx_struct;
#endif

#ifdef SUPPORT_ERROR_PDU_PRINT
INT8U err_clases[7][25] = {
    "ERROR_CLASS_DEVICE",
    "ERROR_CLASS_OBJECT",
    "ERROR_CLASS_PROPERTY",
    "ERROR_CLASS_RESOURCES",
    "ERROR_CLASS_SECURITY",
    "ERROR_CLASS_SERVICES",
    "ERROR_CLASS_VT = 6"
};

INT8U err_codes[51][60] = {
    "ERROR_CODE_OTHER = 0",
    "ERROR_CODE_AUTHENTICATION_FAILED = 1",
    "ERROR_CODE_CONFIGURATION_IN_PROGRESS = 2",
    "ERROR_CODE_DEVICE_BUSY = 3",
    "ERROR_CODE_DYNAMIC_CREATION_NOT_SUPPORTED = 4",
    "ERROR_CODE_FILE_ACCESS_DENIED = 5",
    "ERROR_CODE_INCOMPATIBLE_SECURITY_LEVELS = 6",
    "ERROR_CODE_INCONSISTENT_PARAMETERS = 7",
    "ERROR_CODE_INCONSISTENT_SELECTION_CRITERION = 8",
    "ERROR_CODE_INVALID_DATA_TYPE = 9",
    "ERROR_CODE_INVALID_FILE_ACCESS_METHOD = 10",
    "ERROR_CODE_ERROR_CODE_INVALID_FILE_START_POSITION = 11",
    "ERROR_CODE_INVALID_OPERATOR_NAME = 12",
    "ERROR_CODE_INVALID_PARAMETER_DATA_TYPE = 13",
    "ERROR_CODE_INVALID_TIME_STAMP = 14",
    "ERROR_CODE_KEY_GENERATION_ERROR = 15",
    "ERROR_CODE_MISSING_REQUIRED_PARAMETER = 16",
    "ERROR_CODE_NO_OBJECTS_OF_SPECIFIED_TYPE = 17",
    "ERROR_CODE_NO_SPACE_FOR_OBJECT = 18",
    "ERROR_CODE_NO_SPACE_TO_ADD_LIST_ELEMENT = 19",
    "ERROR_CODE_NO_SPACE_TO_WRITE_PROPERTY = 20",
    "ERROR_CODE_NO_VT_SESSIONS_AVAILABLE = 21",
    "ERROR_CODE_PROPERTY_IS_NOT_A_LIST = 22",
    "ERROR_CODE_OBJECT_DELETION_NOT_PERMITTED = 23",
    "ERROR_CODE_OBJECT_IDENTIFIER_ALREADY_EXISTS = 24",
    "ERROR_CODE_OPERATIONAL_PROBLEM = 25",
    "ERROR_CODE_PASSWORD_FAILURE = 26",
    "ERROR_CODE_READ_ACCESS_DENIED = 27",
    "ERROR_CODE_SECURITY_NOT_SUPPORTED = 28",
    "ERROR_CODE_SERVICE_REQUEST_DENIED = 29",
    "ERROR_CODE_TIMEOUT = 30",
    "ERROR_CODE_UNKNOWN_OBJECT = 31",
    "ERROR_CODE_UNKNOWN_PROPERTY = 32",
    "ERROR_CODE_RESERVED1 = 33",
    "ERROR_CODE_UNKNOWN_VT_CLASS = 34",
    "ERROR_CODE_UNKNOWN_VT_SESSION = 35",
    "ERROR_CODE_UNSUPPORTED_OBJECT_TYPE = 36",
    "ERROR_CODE_VALUE_OUT_OF_RANGE = 37",
    "ERROR_CODE_VT_SESSION_ALREADY_CLOSED = 38",
    "ERROR_CODE_VT_SESSION_TERMINATION_FAILURE = 39",
    "ERROR_CODE_WRITE_ACCESS_DENIED = 40",
    "ERROR_CODE_CHARACTER_SET_NOT_SUPPORTED = 41",
    "ERROR_CODE_INVALID_ARRAY_INDEX = 42",
    "ERROR_CODE_COV_SUBSCRIPTION_FAILED = 43",
    "ERROR_CODE_NOT_COV_PROPERTY = 44",
    "ERROR_CODE_OPTIONAL_FUNCTIONALITY_NOT_SUPPORTED = 45",
    "ERROR_CODE_INVALID_CONFIGURATION_DATA = 46",
    "ERROR_CODE_DATATYPE_NOT_SUPPORTED = 47",
    "ERROR_CODE_DUPLICATE_NAME = 48",
    "ERROR_CODE_DUPLICATE_OBJECT_ID = 49",
    "ERROR_CODE_PROPERTY_IS_NOT_AN_ARRAY = 50"
};

INT8U rej_codes[10][60] = {
    "REJECT_REASON_OTHER = 0",
    "REJECT_REASON_BUFFER_OVERFLOW = 1",
    "REJECT_REASON_INCONSISTENT_PARAMETERS = 2",
    "REJECT_REASON_INVALID_PARAMETER_DATA_TYPE = 3",
    "REJECT_REASON_INVALID_TAG = 4",
    "REJECT_REASON_MISSING_REQUIRED_PARAMETER = 5",
    "REJECT_REASON_PARAMETER_OUT_OF_RANGE = 6",
    "REJECT_REASON_TOO_MANY_ARGUMENTS = 7",
    "REJECT_REASON_UNDEFINED_ENUMERATION = 8",
    "REJECT_REASON_UNRECOGNIZED_SERVICE = 9"
};
#endif
/******* Functions *********************************************************/
/***************************************************************************/
INT8U num_segs_allowed(INT8U MaxS_MaxR)
{
    switch (MaxS_MaxR & APDU_MAX_SEGS_MASK)
    {
        case APDU_MAX_SEGS_2:
            return 2;

        case APDU_MAX_SEGS_4:
            return 4;

        case APDU_MAX_SEGS_8:
            return 8;

        case APDU_MAX_SEGS_16:
            return 16;

        case APDU_MAX_SEGS_32:
            return 32;

        case APDU_MAX_SEGS_64:
            return 64;
    }

    return 255; /* MAX_INT8U */
}

/***************************************************************************/
INT16U segs_size_allowed(INT8U MaxS_MaxR)
{
    switch (MaxS_MaxR & APDU_MAX_LEN_MASK)
    {
        case APDU_MAX_LEN_50:
            return 50;

        case APDU_MAX_LEN_128:
            return 128;

        case APDU_MAX_LEN_206:
            return 206;

        case APDU_MAX_LEN_480:
            return 480;

        case APDU_MAX_LEN_1024:
            return 1024;

        case APDU_MAX_LEN_1476:
            return 1476;
    }

    return 1476; /* RESERVED => ??? */
}


/***************************************************************************/
void reset_appl_layer(struct appl_sruct_t *appl_sruct)
{
    appl_sruct->State = APPL_LAYER_STATE_IDLE;
    appl_sruct->RequestedAPDUSize = MAX_APDU_LEN;
    appl_sruct->SegmentTimer = 0;
    appl_sruct->SequenceNumber = 0;
    appl_sruct->InvokeId = 0;
    appl_sruct->DataLen = 0;

    return;
}

/***************************************************************************/
void init_appl_layer(void)
{
    reset_appl_layer(&appl_rx_struct);

#ifdef SUPPORT_READ_PROPERTY_INIT
    reset_appl_layer(&appl_tx_struct);
#endif

    return;
}

/***************************************************************************/
void APPL_Millisecond_Timer(INT8U dt)
{
    if (appl_rx_struct.State == APPL_LAYER_STATE_SEGM_RESP) {
        if (appl_rx_struct.SegmentTimer < 32768) appl_rx_struct.SegmentTimer+=dt;
        else appl_rx_struct.SegmentTimer=0;

        if (appl_rx_struct.SegmentTimer > APDU_SEGMENT_TIMEOUT) {
            /* segment ack lost */
            no_p_printf("Segment ACK lost!\n");
            reset_appl_layer(&appl_rx_struct);
        }
    }
    return;
}


/***************************************************************************/
BACnet_npdu_t *parse_apdu(BACnet_npdu_t *npdu)
{
    BACnet_npdu_t *resp_pdu;
    INT8U  *apdu;
    INT8U  InvokeId;
    INT8U  allowed_num_segs;
    INT16U i;

    apdu=npdu->data;

    /* PDU Type */
    switch (apdu[0]&APDU_TYPE_MASK)
    {
        case APDU_TYPE_CONFIRMED_REQUEST:
            d_no_p_printf("Received APDU_TYPE_CONFIRMED_REQUEST\n");
            InvokeId = apdu[2];

            if (apdu_is_segmented(apdu[0])==TRUE) {
                no_p_printf((apdu_is_segmented(apdu[0])) ? "segmented request\n": "");
                no_p_printf((apdu_has_more(apdu[0])) ? "apdu_has_more\n" : "");
                no_p_printf((apdu_segm_accept(apdu[0])) ? "apdu_segm_accept\n" : "");
                /* we do not support segmented requests -> send Abort */
                return apdu_abort(npdu, InvokeId, ABORT_REASON_SEGMENTATION_NOT_SUPPORTED);
            }

            /* NOT segmented => NO SeqNum and PropWinSize => Service Choice is at apdu[3] */
            switch (apdu[3])
            {
                case SERVICE_CONFIRMED_READ_PROPERTY:
                    /* check the allowed responce sizes */
                    if (apdu_segm_accept(apdu[0])) {
                        allowed_num_segs = num_segs_allowed(apdu[1]);
                        appl_rx_struct.RequestedAPDUSize = segs_size_allowed(apdu[1]);
                    }
                    else {
                        allowed_num_segs = 1;
                        appl_rx_struct.RequestedAPDUSize = MAX_APDU_LEN;
                    }

                    /* Handle the request */
                    resp_pdu = handle_read_property(npdu, InvokeId);
                    if (resp_pdu == NULL) return NULL;

                    /*  Do we need segmentation to reply? */
                    if (resp_pdu->data_len > appl_rx_struct.RequestedAPDUSize) {

                        /* we can not handle two segmented responces at once */
                        if (appl_rx_struct.State == APPL_LAYER_STATE_SEGM_RESP) {
                            return apdu_abort(npdu, InvokeId, ABORT_REASON_BUFFER_OVERFLOW);
                        }

                        /* are we allowed to send segmented? */
                        if (allowed_num_segs == 1) {
                            return apdu_abort(npdu, InvokeId, ABORT_REASON_BUFFER_OVERFLOW);
                        }

                        /* does the reply fit in the allowed segments */
                        if ((INT32U)(allowed_num_segs * appl_rx_struct.RequestedAPDUSize) < resp_pdu->data_len) {
                            /* are we allowed to send segmented? */
                            if (allowed_num_segs == 1) {
                                return apdu_abort(npdu, InvokeId, ABORT_REASON_BUFFER_OVERFLOW);
                            }
                        }

                        /* OK we can start sending segmented */
                        /* save the length of the data */
                        appl_rx_struct.DataLen=resp_pdu->data_len;

                        no_p_printf("segment this:");
                        /* copy the data to the segmentation buffer */
                        for (i=0; i<resp_pdu->data_len; i++) {
                            appl_rx_struct.Data[i]=resp_pdu->data[i];
                            one_p_printf("%02x ", appl_rx_struct.Data[i]);
                        }
                        no_p_printf("\n\n");
                        appl_rx_struct.SequenceNumber=0;
                        appl_rx_struct.State = APPL_LAYER_STATE_SEGM_RESP;
                        appl_rx_struct.InvokeId = InvokeId;
                        appl_rx_struct.SegmentTimer = 0;

                        /* return the first part */
                        return segmented_c_ack_pdu(npdu,
                                appl_rx_struct.InvokeId,
                                &appl_rx_struct.Data[0],
                                appl_rx_struct.DataLen,
                                appl_rx_struct.RequestedAPDUSize,
                                appl_rx_struct.SequenceNumber,
                                PROPOSED_WS);

                    }
                    else {
                        /* NOT segmented */
                        return resp_pdu;
                    }
                    break;

                case SERVICE_CONFIRMED_WRITE_PROPERTY:
                    return handle_write_property(npdu, InvokeId);

                case SERVICE_CONFIRMED_DEVICE_COMMUNICATION_CONTROL:
                    /*return handle_dev_com_ctrl(npdu, InvokeId);*/
                    /* faltrough to default untill we implement it */
                case SERVICE_CONFIRMED_REINITIALIZE_DEVICE:
                    /*return handle_reinit_dev(npdu, InvokeId);*/
                    /* faltrough to default untill we implement it */
                default :
                    return apdu_reject(npdu, InvokeId, REJECT_REASON_UNRECOGNIZED_SERVICE);
            }
            break;
        case APDU_TYPE_UNCONFIRMED_REQUEST:
            d_no_p_printf("Received APDU_TYPE_UNCONFIRMED_REQUEST\n");
            /* service */
            switch (apdu[1])
            {
                case SERVICE_UNCONFIRMED_WHO_IS:
                    return handle_who_is(npdu);

                case SERVICE_UNCONFIRMED_WHO_HAS:
                    return handle_who_has(npdu);

                case SERVICE_UNCONFIRMED_I_AM:
                    handle_i_am(npdu);
                    break;

                case SERVICE_UNCONFIRMED_I_HAVE:
                    handle_i_have(npdu);
                    break;
            }
            break;
        case APDU_TYPE_SIMPLE_ACK:
            no_p_printf("Received APDU_TYPE_SIMPLE_ACK\n");
            break;

        case APDU_TYPE_COMPLEX_ACK:
            no_p_printf("Received APDU_TYPE_COMPLEX_ACK: ");
            one_p_printf("Invoke ID =%3d ", apdu[1]);

#ifdef SUPPORT_READ_PROPERTY_INIT
            if (apdu_is_segmented(apdu[0])==TRUE) {
                /* is this the first segment? */
                if (apdu[2]==0) {
                    /* this implementation can not handle two segmented transactions at once */
                    if (appl_tx_struct.State == APPL_LAYER_STATE_SEGD_CONF) {
                        return apdu_abort(npdu, InvokeId, ABORT_REASON_BUFFER_OVERFLOW);
                    }
                    else if (appl_tx_struct.DataLen !=0){
                        no_p_printf("appl_tx_struct.DataLen should be 0!!\n");
                    }
                }

                /* copy the apdu data to the local buffer */
                for (i=5; i<npdu->data_len; i++) {
                    appl_tx_struct.Data[appl_tx_struct.DataLen] = npdu->data[i];
                    appl_tx_struct.DataLen++;
                }

                if (!apdu_has_more(apdu[0])) {
                    /* this is the last segment. -> Handle the complete apdu */

                    switch (apdu[4])
                    {
                        case SERVICE_CONFIRMED_READ_PROPERTY:
                            handle_read_propertyACK(appl_tx_struct.Data, appl_tx_struct.DataLen);
                            break;
                        default:
                            no_p_printf("Only SERVICE_CONFIRMED_READ_PROPERTY supported in C ACK\n");
                    }
                    reset_appl_layer(&appl_rx_struct);
                }

                return segment_ack_pdu(npdu, apdu[1], apdu[2], ACTUAL_WS, FALSE, FALSE);
            }
            else {
                /* NOT segmented => NO SeqNum and PropWinSize => Service Choice is at apdu[2] */
                switch (apdu[2])
                {
                    case SERVICE_CONFIRMED_READ_PROPERTY:
                        /* skip unsegmented header => +3*/
                        handle_read_propertyACK(npdu->data+3, npdu->data_len-3);
                        break;
                    default:
                        no_p_printf("Only SERVICE_CONFIRMED_READ_PROPERTY supported in C ACK\n");
                }
            }
#endif
            break;

        case APDU_TYPE_SEGMENT_ACK:
            no_p_printf("Received APDU_TYPE_SEGMENT_ACK\n");
            InvokeId = apdu[1];

            if (appl_rx_struct.State != APPL_LAYER_STATE_SEGM_RESP) {
                /* return error? FIX_ME */
            }
            else {
                /* send next segment */
                appl_rx_struct.SequenceNumber++;
                appl_rx_struct.InvokeId = InvokeId;
                one_p_printf("invokeId=%d\n", InvokeId);
                appl_rx_struct.SegmentTimer = 0;

                resp_pdu = segmented_c_ack_pdu(npdu,
                                               appl_rx_struct.InvokeId,
                                               &appl_rx_struct.Data[0],
                                               appl_rx_struct.DataLen,
                                               appl_rx_struct.RequestedAPDUSize,
                                               appl_rx_struct.SequenceNumber,
                                               ACTUAL_WS);
                if (resp_pdu == NULL) {
                    /* this was an ack for the last segment (or error indication)*/
                    /* reset segmentation information */
                    reset_appl_layer(&appl_rx_struct);
                }
                return resp_pdu;
            }
            break;

        case APDU_TYPE_ERROR:
            no_p_printf("Received APDU_TYPE_ERROR\n");
            one_p_printf("Invoke ID =%3d, ", apdu[1]);
            /* assume read property choise */
#ifdef SUPPORT_ERROR_PDU_PRINT
            two_p_printf("Class = %s, Code = %s\n",err_clases[apdu[4]], err_codes[apdu[6]]);
#endif

#ifdef SUPPORT_READ_PROPERTY_INIT
            if (appl_tx_struct.State == APPL_LAYER_STATE_SEGD_CONF) {
                if (appl_tx_struct.InvokeId == InvokeId) {
                    /* reset segmentation information */
                    reset_appl_layer(&appl_tx_struct);
                }
            }
#endif
            if (appl_rx_struct.State == APPL_LAYER_STATE_SEGM_RESP) {
                if (appl_rx_struct.InvokeId == InvokeId) {
                    /* reset segmentation information */
                    reset_appl_layer(&appl_rx_struct);
                }
            }
            break;

        case APDU_TYPE_REJECT:
            no_p_printf("Received APDU_TYPE_REJECT\n");
#ifdef SUPPORT_ERROR_PDU_PRINT
            one_p_printf("Rejection Cause = %s\n",rej_codes[apdu[2]]);
#endif
#ifdef SUPPORT_READ_PROPERTY_INIT
            if (appl_tx_struct.State == APPL_LAYER_STATE_SEGD_CONF) {
                if (appl_tx_struct.InvokeId == InvokeId) {
                    /* reset segmentation information */
                    reset_appl_layer(&appl_tx_struct);
                }
            }
#endif
            if (appl_rx_struct.State == APPL_LAYER_STATE_SEGM_RESP) {
                if (appl_rx_struct.InvokeId == InvokeId) {
                    /* reset segmentation information */
                    reset_appl_layer(&appl_rx_struct);
                }
            }
            break;

        case APDU_TYPE_ABORT:
            no_p_printf("Received APDU_TYPE_ABORT\n");
#ifdef SUPPORT_READ_PROPERTY_INIT
            if (appl_tx_struct.State == APPL_LAYER_STATE_SEGD_CONF) {
                if (appl_tx_struct.InvokeId == InvokeId) {
                    /* reset segmentation information */
                    reset_appl_layer(&appl_tx_struct);
                }
            }
#endif
            if (appl_rx_struct.State == APPL_LAYER_STATE_SEGM_RESP) {
                if (appl_rx_struct.InvokeId == InvokeId) {
                    /* reset segmentation information */
                    reset_appl_layer(&appl_rx_struct);
                }
            }
            break;
    }
    return NULL;
}
