/*####COPYRIGHTBEGIN####
 -------------------------------------------
 Copyright (C) 2004 Steve Karg

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to:
 The Free Software Foundation, Inc.
 59 Temple Place - Suite 330
 Boston, MA  02111-1307, USA.

 As a special exception, if other files instantiate templates or
 use macros or inline functions from this file, or you compile
 this file and link it with other works to produce a work based
 on this file, this file does not by itself cause the resulting
 work to be covered by the GNU General Public License. However
 the source code for this file must still be made available in
 accordance with section (3) of the GNU General Public License.

 This exception does not invalidate any other reasons why a work
 based on this file might be covered by the GNU General Public
 License.

 Modified by Kåre Särs (kare.sars@wapice.com) March 2005
 -------------------------------------------
####COPYRIGHTEND####*/

#include "platform.h"
#include "conf.h"
#include "bacdcode.h"
#include "bacenum.h"

#ifndef MAX_APDU_LEN
#define MAX_APDU_LEN 480
#endif

#define TAG_CLASS_BIT (0x08)

/**********************************************************************/
INT16U bac_strlen(INT8U *str)
{
    INT8U *s=str;

    while(*str++);
    return (str - s - 1);
}

/* copy four bytes into an 32 bit integer and return the number of bytes used */
/**********************************************************************/
INT8U chars_to_int32(INT8U *ch_str, INT32U *val)
{
    union {
        INT8U byte[4];
        INT32U value;
    } data = {{0}};

#ifdef IS_BIGENDIAN
    data.byte[0] = ch_str[0];
    data.byte[1] = ch_str[1];
    data.byte[2] = ch_str[2];
    data.byte[3] = ch_str[3];
#else
    data.byte[0] = ch_str[3];
    data.byte[1] = ch_str[2];
    data.byte[2] = ch_str[1];
    data.byte[3] = ch_str[0];
#endif
    *val=data.value;
    return 4;
}

/* copy two bytes into an 16 bit integer and return the number of bytes used */
/**********************************************************************/
INT8U chars_to_int16(INT8U *ch_str, INT16U *val)
{
    union {
        INT8U byte[2];
        INT16U value;
    } data = { { 0}};

#ifdef IS_BIGENDIAN
    data.byte[0] = ch_str[0];
    data.byte[1] = ch_str[1];
#else
    data.byte[0] = ch_str[1];
    data.byte[1] = ch_str[0];
#endif
    *val=data.value;
    return 2;
}

/* copy four bytes from an 32 bit integer to an array of chars and */
/* return the number of bytes used */
/**********************************************************************/
INT8U int32_to_chars(INT8U *ch_str, INT32U val)
{
    union {
        INT8U byte[4];
        INT32U value;
    } data;

    data.value=val;

#ifdef IS_BIGENDIAN
    ch_str[0] = data.byte[0];
    ch_str[1] = data.byte[1];
    ch_str[2] = data.byte[2];
    ch_str[3] = data.byte[3];
#else
    ch_str[0] = data.byte[3];
    ch_str[1] = data.byte[2];
    ch_str[2] = data.byte[1];
    ch_str[3] = data.byte[0];
#endif

    return 4;
}

/* copy four bytes from an 16 bit integer to an array of chars and */
/* return the number of bytes used */
/**********************************************************************/
INT8U int16_to_chars(INT8U *ch_str, INT16U val)
{
    union {
        INT8U byte[2];
        INT16U value;
    } data;

    data.value=val;

#ifdef IS_BIGENDIAN
    ch_str[0] = data.byte[0];
    ch_str[1] = data.byte[1];
#else
    ch_str[0] = data.byte[1];
    ch_str[1] = data.byte[0];
#endif

    return 2;
}

/* NOTE: byte order plays a role in decoding multibyte values */
/* http://www.unixpapa.com/incnote/byteorder.html */

/* max-segments-accepted
   B'000'      Unspecified number of segments accepted.
   B'001'      2 segments accepted.
   B'010'      4 segments accepted.
   B'011'      8 segments accepted.
   B'100'      16 segments accepted.
   B'101'      32 segments accepted.
   B'110'      64 segments accepted.
   B'111'      Greater than 64 segments accepted.
*/

/* max-APDU-length-accepted
   B'0000'  Up to MinimumMessageSize (50 octets)
   B'0001'  Up to 128 octets
   B'0010'  Up to 206 octets (fits in a LonTalk frame)
   B'0011'  Up to 480 octets (fits in an ARCNET frame)
   B'0100'  Up to 1024 octets
   B'0101'  Up to 1476 octets (fits in an ISO 8802-3 frame)
   B'0110'  reserved by ASHRAE
   B'0111'  reserved by ASHRAE
   B'1000'  reserved by ASHRAE
   B'1001'  reserved by ASHRAE
   B'1010'  reserved by ASHRAE
   B'1011'  reserved by ASHRAE
   B'1100'  reserved by ASHRAE
   B'1101'  reserved by ASHRAE
   B'1110'  reserved by ASHRAE
   B'1111'  reserved by ASHRAE
*/
/* from clause 20.1.2.4 max-segments-accepted */
/* and clause 20.1.2.5 max-APDU-length-accepted */
/* returns the encoded octet */
/**********************************************************************/
INT8U encode_max_segs_max_apdu(INT8U max_segs, INT16U max_apdu)
{
    INT8U octet = 0;

    if (max_segs < 2)
        octet = 0;
    else if (max_segs < 4)
        octet = 0x10;
    else if (max_segs < 8)
        octet = 0x20;
    else if (max_segs < 16)
        octet = 0x30;
    else if (max_segs < 32)
        octet = 0x40;
    else if (max_segs < 64)
        octet = 0x50;
    else if (max_segs == 64)
        octet = 0x60;
    else
        octet = 0x70;

    /* max_apdu must be 50 octets minimum */
    if (max_apdu <= 50)
        octet |= 0x00;
    else if (max_apdu <= 128)
        octet |= 0x01;
    /*fits in a LonTalk frame */
    else if (max_apdu <= 206)
        octet |= 0x02;
    /*fits in an ARCNET or MS/TP frame */
    else if (max_apdu <= 480)
        octet |= 0x03;
    else if (max_apdu <= 1024)
        octet |= 0x04;
    /* fits in an ISO 8802-3 frame */
    else if (max_apdu <= 1476)
        octet |= 0x05;


    return octet;
}


/* from clause 20.2.1 General Rules for Encoding BACnet Tags */
/* returns the number of apdu bytes consumed */
/**********************************************************************/
INT8U encode_tag(INT8U *apdu, INT8U tag_number, BOOLEAN context_specific, INT32U len_value_type)
{
    INT8U len = 1;                /* return value */

    apdu[0] = 0;
    if (context_specific == TRUE) apdu[0] = TAG_CLASS_BIT;

    /* additional tag byte after this byte */
    /* for extended tag byte */
    if (tag_number <= 14) {
        apdu[0] |= (tag_number << 4);
    }
    else {
        apdu[0] |= 0xF0;
        apdu[1] = tag_number;
        len++;
    }

    /* NOTE: additional len byte(s) after extended tag byte */
    /* if larger than 4 */
    if (len_value_type <= 4)
        apdu[0] |= len_value_type;
    else {
        apdu[0] |= 5;
        if (len_value_type <= 253) {
            apdu[len] = len_value_type&0xff;
            len++;
        } else if (len_value_type <= 65535) {
            apdu[len] = 254;
            len++;
            len += int16_to_chars(&apdu[len], len_value_type&0xffff);
        } else {
            apdu[len] = 255;
            len++;
            len += int32_to_chars(&apdu[len], len_value_type);
        }
    }

    return len;
}

/* from clause 20.2.1.3.2 Constructed Data */
/* returns the number of apdu bytes consumed */
/**********************************************************************/
INT8U encode_opening_tag(INT8U *apdu, INT8U tag_number)
{
    INT8U len = 1;

    apdu[0] = TAG_CLASS_BIT;
    /* additional tag byte after this byte for extended tag byte */
    if (tag_number <= 14)
        apdu[0] |= (tag_number << 4);
    else {
        apdu[0] |= 0xF0;
        apdu[1] = tag_number;
        len++;
    }

    /* type indicates opening tag */
    apdu[0] |= 6;

    return len;
}

/* from clause 20.2.1.3.2 Constructed Data */
/* returns the number of apdu bytes consumed */
/**********************************************************************/
INT8U encode_closing_tag(INT8U *apdu, INT8U tag_number)
{
    INT8U len = 1;

    apdu[0] = TAG_CLASS_BIT;
    /* additional tag byte after this byte for extended tag byte */
    if (tag_number <= 14)
        apdu[0] |= (tag_number << 4);
    else {
        apdu[0] |= 0xF0;
        apdu[1] = tag_number;
        len++;
    }

    /* type indicates closing tag */
    apdu[0] |= 7;

    return len;
}

/**********************************************************************/
static BOOLEAN decode_is_extended_tag_number(INT8U *apdu)
{
    return ((apdu[0] & 0xF0) == 0xF0) ? TRUE : FALSE;
}

/**********************************************************************/
static BOOLEAN decode_is_extended_value(INT8U *apdu)
{
    return ((apdu[0] & 0x07) == 5) ? TRUE : FALSE;
}

/**********************************************************************/
BOOLEAN decode_is_context_specific(INT8U *apdu)
{
    return ((apdu[0] & TAG_CLASS_BIT) != 0) ? TRUE : FALSE;
}

/**********************************************************************/
static BOOLEAN decode_is_opening_tag(INT8U *apdu)
{
    return ((apdu[0] & 0x07) == 6) ? TRUE : FALSE;
}

/**********************************************************************/
static BOOLEAN decode_is_closing_tag(INT8U *apdu)
{
    return ((apdu[0] & 0x07) == 7) ? TRUE : FALSE;
}

/**********************************************************************/
INT8U decode_tag_number(INT8U *apdu, INT8U *tag_number)
{
    INT8U len = 1; /* return value */

    /* decode the tag number first */
    if (decode_is_extended_tag_number(&apdu[0])) {
        /* extended tag */
        if (tag_number != NULL) *tag_number = apdu[1];
        len++;
    } else {
        if (tag_number != NULL) *tag_number = (apdu[0] >> 4);
    }

    return len;
}

/* from clause 20.2.1.3.2 Constructed Data */
/* returns the number of apdu bytes consumed */
/**********************************************************************/
INT8U decode_tag_number_and_lvt(INT8U *apdu, INT8U *tag_number, INT32U *value)
{
    INT8U  len = 1;
    INT16U int16v;
    INT32U int32v;

    len = decode_tag_number(&apdu[0], tag_number);
    /* decode the value */
    if (decode_is_extended_value(&apdu[0])) {
        /* tagged as INT32U */
        if (apdu[len] == 255) {
            len++;
            len += chars_to_int32(&apdu[len], &int32v);
            if (value != NULL) *value=int32v;
        }
        /* tagged as INT16U */
        else if (apdu[len] == 254) {
            len++;
            len += chars_to_int16(&apdu[len], &int16v);
            if (value != NULL) *value=int16v;
        }
        /* no tag - must be INT8U */
        else {
            if (value != NULL) *value = apdu[len];
            len++;
        }
    }
    else if (decode_is_opening_tag(&apdu[0]) && value) *value = 0;
    /* closing tag */
    else if (decode_is_closing_tag(&apdu[0]) && value) *value = 0;
    /* small value */
    else if (value != NULL) *value = apdu[0] & 0x07;

    return len;
}

/* from clause 20.2.1.3.2 Constructed Data */
/* returns true if the tag is context specific and matches */
/**********************************************************************/
BOOLEAN decode_is_context_tag(INT8U *apdu, INT8U tag_number)
{
    INT8U my_tag_number = 0;
    BOOLEAN context_specific = FALSE;

    context_specific = decode_is_context_specific(apdu);
    (void)decode_tag_number(apdu,&my_tag_number);

    return ((context_specific == TRUE) && (my_tag_number == tag_number)) ? TRUE : FALSE;
}

/* from clause 20.2.1.3.2 Constructed Data */
/* returns the true if the tag is an opening tag and the tag number matches */
/**********************************************************************/
BOOLEAN decode_is_opening_tag_number(INT8U *apdu, INT8U tag_number)
{
    INT8U my_tag_number = 0;
    BOOLEAN opening_tag = FALSE;

    opening_tag = decode_is_opening_tag(apdu);
    (void)decode_tag_number(apdu,&my_tag_number);

    return ((opening_tag == TRUE) && (my_tag_number == tag_number)) ? TRUE : FALSE;
}

/* from clause 20.2.1.3.2 Constructed Data */
/* returns the true if the tag is a closing tag and the tag number matches */
/**********************************************************************/
BOOLEAN decode_is_closing_tag_number(INT8U *apdu, INT8U tag_number)
{
    INT8U my_tag_number = 0;
    BOOLEAN closing_tag = FALSE;

    closing_tag = decode_is_closing_tag(apdu);
    (void)decode_tag_number(apdu,&my_tag_number);

    return ((closing_tag == TRUE) && (my_tag_number == tag_number)) ? TRUE : FALSE;
}


/* from clause 20.2.2 Encoding of Null */
/* returns the number of apdu bytes consumed */
/**********************************************************************/
INT16U encode_tagged_null(INT8U *apdu)
{
    apdu[0]=0;
    return 1;
}


/* from clause 20.2.3 Encoding of a Boolean */
/* returns the number of apdu bytes consumed */
/**********************************************************************/
INT16U encode_tagged_boolean(INT8U *apdu, BOOLEAN value)
{
    return encode_tag(&apdu[0], BACNET_APPLICATION_TAG_BOOLEAN, FALSE, ((value == TRUE) ? 1 : 0));
}

/**********************************************************************/
INT16U encode_context_boolean(INT8U *apdu, INT8U tag_number, BOOLEAN value)
{
    INT16U len;

    len = encode_tag(&apdu[0], tag_number, TRUE, 1);
    apdu[len] = ((value == TRUE) ? 1 : 0);
    len ++;
    return len;
}


/* returns the number of apdu bytes consumed 0 on failure */
/**********************************************************************/
INT16U decode_tagged_boolean(INT8U *apdu, BOOLEAN *value)
{
    /* check that it is not a context_specific tag */
    if (decode_is_context_specific(&apdu[0]) == TRUE) return 0;

    /* check that it is a BACNET_APPLICATION_TAG_BOOLEAN tag */
    if ((apdu[0]&0xf0) != 0x10) return 0; /* 0x10 == BACNET_APPLICATION_TAG_BOOLEAN << 4 */

    /* value == True if 1 and false if 0 */
    *value = ((apdu[0]&0x01) == 0x1) ? TRUE : FALSE;

    return 1; /* a tagged boolean takes only one byte */
}

/* from clause 20.2.6 Encoding of a Real Number Value */
/* returns the number of apdu bytes consumed */
/**********************************************************************/
INT8U chars_to_real(INT8U *apdu, REAL *real_value)
{
    union {
        INT8U byte[4];
        REAL real_value;
    } my_data;

#ifdef IS_BIGENDIAN
    my_data.byte[0] = apdu[0];
    my_data.byte[1] = apdu[1];
    my_data.byte[2] = apdu[2];
    my_data.byte[3] = apdu[3];
#else
    my_data.byte[0] = apdu[3];
    my_data.byte[1] = apdu[2];
    my_data.byte[2] = apdu[1];
    my_data.byte[3] = apdu[0];
#endif

    *real_value = my_data.real_value;

    return 4;
}

/* returns the number of apdu bytes consumed, 0 on error */
/**********************************************************************/
INT16U decode_tagged_real(INT8U *apdu, REAL *real_value)
{
    INT8U tag;
    INT16U len;

    /* context specific == error */
    if (decode_is_context_specific(apdu) == TRUE) return 0;

    /* tag number must be BACNET_APPLICATION_TAG_REAL */
    len = decode_tag_number(apdu, &tag);
    if (tag != BACNET_APPLICATION_TAG_REAL) return 0;

    len += chars_to_real(&apdu[len], real_value);

    return len;
}


/* from clause 20.2.6 Encoding of a Real Number Value */
/* returns the number of apdu bytes consumed */
/**********************************************************************/
INT8U real_to_chars(INT8U *apdu, REAL value)
{
    union {
        INT8U byte[4];
        REAL real_value;
    } my_data;

    my_data.real_value = value;

#ifdef IS_BIGENDIAN
    apdu[0] = my_data.byte[0];
    apdu[1] = my_data.byte[1];
    apdu[2] = my_data.byte[2];
    apdu[3] = my_data.byte[3];
#else
    apdu[0] = my_data.byte[3];
    apdu[1] = my_data.byte[2];
    apdu[2] = my_data.byte[1];
    apdu[3] = my_data.byte[0];
#endif

    return 4;
}

/* from clause 20.2.6 Encoding of a Real Number Value */
/* and 20.2.1 General Rules for Encoding BACnet Tags */
/* returns the number of apdu bytes consumed */
/**********************************************************************/
INT16U encode_tagged_real(INT8U *apdu, REAL value)
{
    INT16U len = 0;

    /* assumes that the tag only consumes 1 octet */
    len = real_to_chars(&apdu[1], value);
    len += encode_tag(&apdu[0], BACNET_APPLICATION_TAG_REAL, FALSE, len);

    return len;
}

/**********************************************************************/
INT16U encode_context_real(INT8U *apdu, INT8U tag_number, REAL value)
{
    INT16U len = 0;

    /* assumes that the tag only consumes 1 octet */
    len = real_to_chars(&apdu[1], value);
    len += encode_tag(&apdu[0], tag_number, TRUE, len);

    return len;
}

/* from clause 20.2.14 Encoding of an Object Identifier Value */
/* returns the number of apdu bytes consumed */
/**********************************************************************/
INT16U encode_bacnet_object_id(INT8U *apdu, INT16U object_type, INT32U instance)
{
    return int32_to_chars(apdu, ((INT32U)(object_type & 0x3FF) << 22) | (instance & 0x3FFFFF));
}

/* from clause 20.2.14 Encoding of an Object Identifier Value */
/* and 20.2.1 General Rules for Encoding BACnet Tags */
/* returns the number of apdu bytes consumed */
/**********************************************************************/
INT16U encode_context_object_id(INT8U *apdu, INT8U tag_number, INT16U object_type, INT32U instance)
{
    INT16U len = 0;

    /* assumes that the tag only consumes 1 octet */
    len = encode_bacnet_object_id(&apdu[1], object_type, instance);
    len += encode_tag(&apdu[0], tag_number, TRUE, len);

    return len;
}

/* from clause 20.2.14 Encoding of an Object Identifier Value */
/* and 20.2.1 General Rules for Encoding BACnet Tags */
/* returns the number of apdu bytes consumed */
/**********************************************************************/
INT16U encode_tagged_object_id(INT8U *apdu, INT16U object_type, INT32U instance)
{
    INT16U len = 0;

    /* assumes that the tag only consumes 1 octet */
    len = encode_bacnet_object_id(&apdu[1], object_type, instance);
    len += encode_tag(&apdu[0], BACNET_APPLICATION_TAG_OBJECT_ID, FALSE, len);

    return len;
}


/* from clause 20.2.14 Encoding of an Object Identifier Value */
/* returns the number of apdu bytes consumed */
/**********************************************************************/
INT16U decode_object_id(INT8U *apdu, INT16U *object_type, INT32U *instance)
{
    INT32U int32v;

    (void)chars_to_int32(apdu, &int32v); /* returns 4 */

    *object_type = ((int32v >> 22) & 0x3FF);
    *instance = (int32v & 0x3FFFFF);

    return 4;
}


/**********************************************************************/
INT16U decode_context_object_id(INT8U *apdu, INT8U tag_number, INT16U *object_type, INT32U *instance)
{
    INT16U len = 0;                /* return value */
    INT32U len_value = 0;
    INT8U tag_num = 0;

    len = decode_tag_number_and_lvt(&apdu[0], &tag_num, &len_value);
    if (tag_num != tag_number) return 0; /* this is not a the expected tag */

    len += decode_object_id(&apdu[len], object_type, instance);

    return len;
}
/**********************************************************************/
INT16U decode_tagged_object_id(INT8U *apdu, INT16U *object_type, INT32U *instance)
{
    INT16U len = 0;                /* return value */
    INT32U len_value = 0;
    INT8U tag_num = 0;

    len = decode_tag_number_and_lvt(&apdu[0], &tag_num, &len_value);
    if (tag_num != BACNET_APPLICATION_TAG_OBJECT_ID) return 0; /* this is not a the expected tag */

    len += decode_object_id(&apdu[len], object_type, instance);

    return len;
}


/* from clause 20.2.9 Encoding of a Character String Value */
/* returns the number of apdu bytes consumed */
/**********************************************************************/
INT16U encode_bacnet_string(INT8U *apdu, INT8U *char_string, INT16U len)
{
    INT16U i;

    /* limit - 6 octets is the most our tag and type could be */
    if (len > (MAX_APDU_LEN - 6)) len = MAX_APDU_LEN - 6;

    for (i = 0; i < len; i++) {
        apdu[1 + i] = char_string[i];
    }
    apdu[0] = CHARACTER_ANSI;
    len++;

    return len;
}

/* from clause 20.2.9 Encoding of a Character String Value */
/* returns the number of apdu bytes consumed */
/**********************************************************************/
INT16U encode_bacnet_char_string(INT8U *apdu, INT8U *char_string)
{
    INT16U len;

    len = bac_strlen(char_string);
    len = encode_bacnet_string(&apdu[0], char_string, len);

    return len;
}

/* from clause 20.2.9 Encoding of a Character String Value */
/* returns the number of apdu bytes consumed */
/**********************************************************************/
INT16U encode_context_char_string(INT8U *apdu, INT8U tag_number, INT8U *char_string)
{
    INT16U len;
    INT16U string_len;

    string_len = bac_strlen(char_string);

    /* An extra type byte is added to the beginning of an ANSI X3.4 string */
    len = encode_tag(&apdu[0], tag_number, TRUE, string_len + 1);
    /* assert((len + string_len) < MAX_APDU_LEN); */
    len += encode_bacnet_string(&apdu[len], char_string, string_len);

    return len;
}

/**********************************************************************/
INT16U encode_tagged_char_string(INT8U *apdu, INT8U *char_string)
{
    INT16U len;
    INT16U string_len;

    string_len = bac_strlen(char_string);

    /* An extra type byte is added to the beginning of an ANSI X3.4 string */
    len = encode_tag(&apdu[0], BACNET_APPLICATION_TAG_CHARACTER_STRING, FALSE, string_len + 1);
    /* assert((len + string_len) < MAX_APDU_LEN); */
    len += encode_bacnet_string(&apdu[len], char_string, string_len);

    return len;
}

/* from clause 20.2.9 Encoding of a Character String Value */
/* and 20.2.1 General Rules for Encoding BACnet Tags */
/* returns the number of apdu bytes consumed */
/**********************************************************************/
INT16U decode_context_char_string(INT8U *apdu, INT8U tag_number, INT8U *char_string, INT16U str_len)
{
    INT16U len = 0, tmp_len;
    INT32U i = 0;
    INT32U len_value = 0;
    INT8U tag_num = 0;

    len = decode_tag_number_and_lvt(&apdu[0], &tag_num, &len_value);
    if (tag_num != tag_number) return 0; /* this is not a the expected tag */
    if (len_value > 0xFFFF) return 0; /* we do not support this long strings */

    if (len_value != 0) {
        /* decode ANSI X3.4 */
        if (apdu[len] == 0) {
            len++;       /* the one type octet */
            len_value--; /* the one type octet */
            tmp_len = (len_value < str_len) ? (INT16U)(len_value&0xFFFF) : str_len-1;
            for (i = 0; i < tmp_len; i++) {
                char_string[i] = apdu[len + i];
            }
            /* terminate the c string */
            char_string[i] = 0;
        }
        len += (INT16U)(len_value&0xFFFF);
    }

    return len;
}

/**********************************************************************/
INT16U decode_tagged_char_string(INT8U *apdu, INT8U *char_string, INT16U str_len)
{
    INT16U len = 0, tmp_len;
    INT32U i = 0;
    INT32U len_value = 0;
    INT8U tag_number = 0;

    len = decode_tag_number_and_lvt(&apdu[0], &tag_number, &len_value);
    /* check tag number */
    if (tag_number != BACNET_APPLICATION_TAG_CHARACTER_STRING) return 0;
    if (len_value > 0xFFFF) return 0; /* we do not support this long strings */

    if (len_value != 0) {
        /* decode ANSI X3.4 */
        if (apdu[len] == 0) {
            len++;       /* the one type octet */
            len_value--; /* the one type octet */
            tmp_len = (len_value < str_len) ? (INT16U)(len_value&0xFFFF) : str_len-1;
            for (i = 0; i < tmp_len; i++) {
                char_string[i] = apdu[len + i];
            }
            /* terminate the c string */
            char_string[i] = 0;
        }
    }
    len += (INT16U)(len_value&0xFFFF);

    return len;
}

/* This function is NOT SAFE!!! */
/* - The string shares memory with the apdu */
/* - No string termination */
/* + Saves memory */
/**********************************************************************/
INT16U decode_context_char_str_ptr(INT8U *apdu, INT8U tag_number, INT8U **char_string, INT16U *str_len)
{
    INT16U len = 0;
    INT32U len_value = 0;
    INT8U tag_num = 0;

    len = decode_tag_number_and_lvt(&apdu[0], &tag_num, &len_value);
    if (tag_num != tag_number) return 0; /* this is not the expected tag */
    if (len_value > 0xFFFF) return 0; /* we do not support this long strings */

    if (len_value != 0) {
        /* decode ANSI X3.4 */
        if (apdu[len] == 0) {
            len++;
            len_value--;
            *char_string = &apdu[len];
            *str_len = (INT16U)(len_value&0xFFFF);
        }
        /* Unsupported string format */
        else {
            *char_string = NULL;
            return 0;
        }
    }

    return (INT16U)(len_value&0xFFFF)+len;
}

/* This function is NOT SAFE!!! */
/* - The string shares memory with the apdu */
/* - No string termination */
/* + Saves memory */
/**********************************************************************/
INT16U decode_tagged_char_str_ptr(INT8U *apdu, INT8U **char_string, INT16U *str_len)
{
    INT16U len = 0;
    INT32U len_value = 0;
    INT8U tag_number = 0;

    len = decode_tag_number_and_lvt(&apdu[0], &tag_number, &len_value);
    if (tag_number != BACNET_APPLICATION_TAG_CHARACTER_STRING) return 0; /* this is not the expected tag */
    if (len_value > 0xFFFF) return 0; /* we do not support this long strings */

    if (len_value != 0) {
        /* decode ANSI X3.4 */
        if (apdu[len] == 0) {
            len++;
            len_value--;
            *char_string = &apdu[len];
            *str_len = (INT16U)(len_value&0xFFFF);
        }
        /* Unsupported string format */
        else {
            *char_string = NULL;
            return 0;
        }
    }

    return (INT16U)(len_value&0xFFFF)+len;
}

/* from clause 20.2.10 Encoding of a Bit String Value */
/* returns the number of apdu bytes consumed */
/* calculate the length of the bitstring and the empty bits at the end */
/**********************************************************************/
INT16U encode_tagged_bit_string(INT8U *apdu, INT8U *bit_string, INT8U last_bit)
{
    INT16U len, i, strlen;

    strlen = (last_bit / 8) + 1;

    /* an inital octet contains the number of unused bits in the final octet */
    len = encode_tag(&apdu[0], BACNET_APPLICATION_TAG_BIT_STRING, FALSE, strlen+1);

    apdu[len] = (strlen * 8) - last_bit - 1;

    for (i=0; i<strlen; i++)
    {
        apdu[i+len+1]=bit_string[i];
    }
    len += i+1;

    return len;
}

/* set the specifyed bit to 1 if TRUE or 0 if FALSE */
/**********************************************************************/
void set_bit_in_string(INT8U *apdu, INT8U bit, BOOLEAN enabeled)
{
    INT8U byte, bbit, mask;
    /* find byte to modify */
    byte = (INT8U)(bit / 8);

    /* find bit to change (bit 0 is the most significant) */
    bbit = ((byte + 1) * 8) - bit - 1;

    /* generate the mask */
    mask = 1 << bbit;

    if (enabeled==TRUE)
    {
        /* set the bit */
        apdu[byte] |= mask;
    }
    else
    {
        /* clear the bit */
        apdu[byte] &= ~mask;
    }
    return;
}


/* from clause 20.2.4 Encoding of an Unsigned Integer Value */
/* and 20.2.1 General Rules for Encoding BACnet Tags */
/* returns the number of apdu bytes consumed */
/**********************************************************************/
INT8U encode_bacnet_unsigned(INT8U *apdu, INT32U value)
{
    union {
        INT8U byte[4];
        INT32U value;
    } data;

    data.value=value;

#ifdef IS_BIGENDIAN
    if (value < 0x100) {
        apdu[0] = data.byte[3];
        return 1;
    } else if (value < 0x10000) {
        apdu[0] = data.byte[2];
        apdu[1] = data.byte[3];
        return 2;
    } else if (value < 0x1000000) {
        apdu[0] = data.byte[1];
        apdu[1] = data.byte[2];
        apdu[2] = data.byte[3];
        return 3;
    } else {
        apdu[0] = data.byte[0];
        apdu[1] = data.byte[1];
        apdu[2] = data.byte[2];
        apdu[3] = data.byte[3];
        return 4;
    }
#else
    if (value < 0x100) {
        apdu[0] = data.byte[0];
        return 1;
    } else if (value < 0x10000) {
        apdu[0] = data.byte[1];
        apdu[1] = data.byte[0];
        return 2;
    } else if (value < 0x1000000) {
        apdu[0] = data.byte[2];
        apdu[1] = data.byte[1];
        apdu[2] = data.byte[0];
        return 3;
    } else {
        apdu[0] = data.byte[3];
        apdu[1] = data.byte[2];
        apdu[2] = data.byte[1];
        apdu[3] = data.byte[0];
        return 4;
    }
#endif
}

/* from clause 20.2.4 Encoding of an Unsigned Integer Value */
/* and 20.2.1 General Rules for Encoding BACnet Tags */
/* returns the number of apdu bytes consumed */
/**********************************************************************/
INT8U encode_context_unsigned(INT8U *apdu, INT8U tag_number, INT32U value)
{
    INT8U len = 0;

    len = encode_bacnet_unsigned(&apdu[1], value);
    len += encode_tag(&apdu[0], tag_number, TRUE, len);

    return len;
}

/* from clause 20.2.4 Encoding of an Unsigned Integer Value */
/* and 20.2.1 General Rules for Encoding BACnet Tags */
/* returns the number of apdu bytes consumed */
/**********************************************************************/
INT8U encode_tagged_unsigned(INT8U *apdu, INT32U value)
{
    INT8U len = 0;

    len = encode_bacnet_unsigned(&apdu[1], value);
    len += encode_tag(&apdu[0], BACNET_APPLICATION_TAG_UNSIGNED_INT, FALSE, len);

    return len;
}

/* from clause 20.2.4 Encoding of an Unsigned Integer Value */
/* and 20.2.1 General Rules for Encoding BACnet Tags */
/* returns the number of apdu bytes consumed. (0==error) */
/**********************************************************************/
INT8U decode_tagged_unsigned(INT8U *apdu, INT32U *value)
{
    INT8U len = 0;                /* return value */
    INT32U len_value = 0;
    INT8U tag_number = 0;

    len = decode_tag_number_and_lvt(&apdu[0], &tag_number, &len_value);
    if (tag_number != BACNET_APPLICATION_TAG_UNSIGNED_INT) return 0;


    switch (len_value) {
        case 1:
            *value = apdu[len];
            break;
        case 2:
            *value  = (INT32U)apdu[len+0]<<8;
            *value += (INT32U)apdu[len+1];
            break;
        case 3:
            *value  = (INT32U)apdu[len+0]<<16;
            *value += (INT32U)apdu[len+1]<<8;
            *value += (INT32U)apdu[len+2];
            break;
        case 4:
            *value  = (INT32U)apdu[len+0]<<24;
            *value  = (INT32U)apdu[len+1]<<16;
            *value += (INT32U)apdu[len+2]<<8;
            *value += (INT32U)apdu[len+3];
            break;
        default:
            *value = 0;
            break;
    }
    len += (INT8U)(len_value&0xFF);

    return len;
}

/* returns the number of apdu bytes consumed */
/* IF it is a context_specific unsigned value */
/* on error return 0 */
/**********************************************************************/
INT8U decode_context_unsigned(INT8U *apdu, INT8U tag_number, INT32U *value)
{
    INT8U i=0;
    INT8U  tag_num=0;
    INT32U  len=0;

    /* context_specific */
    if (decode_is_context_tag(apdu, tag_number) == FALSE) return 0;

    /* get the tag number and data length */
    i += decode_tag_number_and_lvt(&apdu[i], &tag_num, &len);
    if (tag_num != tag_number) return 0;

    /* get the value */
    switch (len) {
        case 1:
            *value = apdu[i];
            break;
        case 2:
            *value  = (INT32U)apdu[i+0]<<8;
            *value += (INT32U)apdu[i+1];
            break;
        case 3:
            *value  = (INT32U)apdu[i+0]<<16;
            *value += (INT32U)apdu[i+1]<<8;
            *value += (INT32U)apdu[i+2];
            break;
        case 4:
            *value  = (INT32U)apdu[i+0]<<24;
            *value  = (INT32U)apdu[i+1]<<16;
            *value += (INT32U)apdu[i+2]<<8;
            *value += (INT32U)apdu[i+3];
            break;
        default:
            *value = 0;
            break;
    }
    i += (INT8U)len;


    return i;
}

/**********************************************************************/
INT8U decode_context_unsigned8(INT8U *apdu, INT8U tag_number, INT8U *value)
{
    INT8U  tag_num=0;
    INT8U  len=0;

    /* context_specific */
    if (decode_is_context_tag(apdu, tag_number) == FALSE) return 0;

    /* get the tag number and data length */
    len = decode_tag_number(&apdu[0], &tag_num);

    if (len != 1) return 0;  /* tag must be one octet */

    if (tag_num != tag_number) return 0;

    if ((apdu[0]&0x07) != 1) return 0; /* only one octet long payload accepted */

    *value = apdu[1];

    return 2;
}


/* from clause 20.2.11 Encoding of an Enumerated Value */
/* and 20.2.1 General Rules for Encoding BACnet Tags */
/* returns the number of apdu bytes consumed */
/**********************************************************************/
INT16U decode_tagged_enum(INT8U *apdu, INT32U *value)
{
    INT16U len = 0;                /* return value */
    INT32U len_value = 0;
    INT8U tag_number = 0;

    len = decode_tag_number_and_lvt(&apdu[0], &tag_number, &len_value);
    if (tag_number != BACNET_APPLICATION_TAG_ENUMERATED) return 0;


    switch (len_value) {
        case 1:
            *value = apdu[len];
            break;
        case 2:
            *value  = (INT32U)apdu[len+0]<<8;
            *value += (INT32U)apdu[len+1];
            break;
        case 3:
            *value  = (INT32U)apdu[len+0]<<16;
            *value += (INT32U)apdu[len+1]<<8;
            *value += (INT32U)apdu[len+2];
            break;
        case 4:
            *value  = (INT32U)apdu[len+0]<<24;
            *value  = (INT32U)apdu[len+1]<<16;
            *value += (INT32U)apdu[len+2]<<8;
            *value += (INT32U)apdu[len+3];
            break;
        default:
            *value = 0;
            break;
    }
    len += (INT16U)len_value;

    return len;
}

/**********************************************************************/
INT16U decode_context_enum(INT8U *apdu, INT8U tag_number, INT32U *value)
{
    return decode_context_unsigned(apdu, tag_number, value);
}


/* from clause 20.2.11 Encoding of an Enumerated Value */
/* and 20.2.1 General Rules for Encoding BACnet Tags */
/* returns the number of apdu bytes consumed */
/**********************************************************************/
INT16U encode_bacnet_enum(INT8U *apdu, INT32U value)
{
    return encode_bacnet_unsigned(apdu, value);
}

/* from clause 20.2.11 Encoding of an Enumerated Value */
/* and 20.2.1 General Rules for Encoding BACnet Tags */
/* returns the number of apdu bytes consumed */
/**********************************************************************/
INT16U encode_tagged_enum(INT8U *apdu, INT32U value)
{
    INT16U len = 0;                /* return value */

    /* assumes that the tag only consumes 1 octet */
    len = encode_bacnet_enum(&apdu[1], value);
    len += encode_tag(&apdu[0], BACNET_APPLICATION_TAG_ENUMERATED, FALSE, len);

    return len;
}

/* from clause 20.2.11 Encoding of an Enumerated Value */
/* and 20.2.1 General Rules for Encoding BACnet Tags */
/* returns the number of apdu bytes consumed */
/**********************************************************************/
INT16U encode_context_enum(INT8U *apdu, INT8U tag_number, INT32U value)
{
    INT16U len = 0;                /* return value */

    /* assumes that the tag only consumes 1 octet */
    len = encode_bacnet_enum(&apdu[1], value);
    len += encode_tag(&apdu[0], tag_number, TRUE, len);

    return len;
}

/* from clause 20.2.5 Encoding of a Signed Integer Value */
/* and 20.2.1 General Rules for Encoding BACnet Tags */
/* returns the number of apdu bytes consumed */
/**********************************************************************/
INT16U decode_tagged_signed(INT8U *apdu, INT32S *value)
{
    return decode_tagged_unsigned(apdu, (INT32U *)value);
}

/**********************************************************************/
INT16U decode_context_signed(INT8U *apdu, INT8U tag_number, INT32S *value)
{
    return decode_context_unsigned(apdu, tag_number, (INT32U *)value);
}

/* from clause 20.2.5 Encoding of a Signed Integer Value */
/* and 20.2.1 General Rules for Encoding BACnet Tags */
/* returns the number of apdu bytes consumed */
/**********************************************************************/
INT16U encode_bacnet_signed(INT8U *apdu, INT32S value)
{
    return encode_bacnet_unsigned(apdu, value);
}

/* from clause 20.2.5 Encoding of a Signed Integer Value */
/* and 20.2.1 General Rules for Encoding BACnet Tags */
/* returns the number of apdu bytes consumed */
/**********************************************************************/
INT16U encode_tagged_signed(INT8U *apdu, INT32S value)
{
    INT16U len = 0;                /* return value */

    /* assumes that the tag only consumes 1 octet */
    len = encode_bacnet_signed(&apdu[1], value);
    len += encode_tag(&apdu[0], BACNET_APPLICATION_TAG_SIGNED_INT, FALSE, len);

    return len;
}

/* from clause 20.2.5 Encoding of a Signed Integer Value */
/* and 20.2.1 General Rules for Encoding BACnet Tags */
/* returns the number of apdu bytes consumed */
/**********************************************************************/
INT16U encode_context_signed(INT8U *apdu, INT8U tag_number, INT32S value)
{
    INT16U len = 0;                /* return value */

    /* assumes that the tag only consumes 1 octet */
    len = encode_bacnet_signed(&apdu[1], value);
    len += encode_tag(&apdu[0], tag_number, TRUE, len);

    return len;
}

/* from clause 20.2.13 Encoding of a Time Value */
/* and 20.2.1 General Rules for Encoding BACnet Tags */
/* returns the number of apdu bytes consumed */
/**********************************************************************/
INT16U encode_bacnet_time(INT8U *apdu, INT8U hour, INT8U min, INT8U sec, INT8U hundredths)
{
    apdu[0] = hour;
    apdu[1] = min;
    apdu[2] = sec;
    apdu[3] = hundredths;

    return 4;
}

/* from clause 20.2.13 Encoding of a Time Value */
/* and 20.2.1 General Rules for Encoding BACnet Tags */
/* returns the number of apdu bytes consumed */
/**********************************************************************/
INT16U encode_tagged_time(INT8U *apdu, INT8U hour, INT8U min, INT8U sec, INT8U hundredths)
{
    INT16U len = 0;

    /* assumes that the tag only consumes 1 octet */
    len = encode_bacnet_time(&apdu[1], hour, min, sec, hundredths);
    len += encode_tag(&apdu[0], BACNET_APPLICATION_TAG_TIME, FALSE, len);

    return len;

}

/* from clause 20.2.13 Encoding of a Time Value */
/* and 20.2.1 General Rules for Encoding BACnet Tags */
/* returns the number of apdu bytes consumed */
/**********************************************************************/
INT16U decode_bacnet_time(INT8U *apdu, INT8U *hour, INT8U *min, INT8U *sec, INT8U *hundredths)
{
    *hour = apdu[0];
    *min = apdu[1];
    *sec = apdu[2];
    *hundredths = apdu[3];

    return 4;
}

/* BACnet Date */
/* year = years since 1900 */
/* month 1=Jan */
/* day = day of month */
/* wday 1=Monday...7=Sunday */

/* from clause 20.2.12 Encoding of a Date Value */
/* and 20.2.1 General Rules for Encoding BACnet Tags */
/* returns the number of apdu bytes consumed */
/**********************************************************************/
INT16U encode_bacnet_date(INT8U *apdu, INT8U year, INT8U month, INT8U day, INT8U wday)
{
    apdu[0] = year;
    apdu[1] = month;
    apdu[2] = day;
    apdu[3] = wday;

    return 4;
}

/* from clause 20.2.12 Encoding of a Date Value */
        /* and 20.2.1 General Rules for Encoding BACnet Tags */
/* returns the number of apdu bytes consumed */
/**********************************************************************/
INT16U encode_tagged_date(INT8U *apdu, INT8U year, INT8U month, INT8U day, INT8U wday)
{
    INT16U len = 0;

    /* assumes that the tag only consumes 1 octet */
    len = encode_bacnet_date(&apdu[1], year, month, day, wday);
    len += encode_tag(&apdu[0], BACNET_APPLICATION_TAG_DATE, FALSE, len);

    return len;

}

/* from clause 20.2.12 Encoding of a Date Value */
/* and 20.2.1 General Rules for Encoding BACnet Tags */
/* returns the number of apdu bytes consumed */
/**********************************************************************/
INT16U decode_date(INT8U *apdu, INT8U *year, INT8U *month, INT8U *day, INT8U *wday)
{
    *year = apdu[0];
    *month = apdu[1];
    *day = apdu[2];
    *wday = apdu[3];

    return 4;
}

