/*####COPYRIGHTBEGIN####
 -------------------------------------------
 Copyright (C) 2004 Steve Karg

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to:
 The Free Software Foundation, Inc.
 59 Temple Place - Suite 330
 Boston, MA  02111-1307, USA.

 As a special exception, if other files instantiate templates or
 use macros or inline functions from this file, or you compile
 this file and link it with other works to produce a work based
 on this file, this file does not by itself cause the resulting
 work to be covered by the GNU General Public License. However
 the source code for this file must still be made available in
 accordance with section (3) of the GNU General Public License.

 This exception does not invalidate any other reasons why a work
 based on this file might be covered by the GNU General Public
 License.

 Modified by Kåre Särs (kare.sars@wapice.com) March 2005
 -------------------------------------------
####COPYRIGHTEND####*/
#ifndef BACDCODE_H
#define BACDCODE_H

/* from clause 20.1.2.4 max-segments-accepted */
/* and clause 20.1.2.5 max-APDU-length-accepted */
/* returns the encoded octet */
INT8U encode_max_segs_max_apdu(INT8U max_segs, INT16U max_apdu);

/* from clause 20.2.1 General Rules for Encoding BACnet Tags */
/* returns the number of apdu bytes consumed */
INT8U encode_tag(INT8U *apdu, INT8U tag_number, BOOLEAN context_specific, INT32U len_value_type);

/* from clause 20.2.1.3.2 Constructed Data */
/* returns the number of apdu bytes consumed */
INT8U encode_opening_tag(INT8U *apdu, INT8U tag_number);
INT8U encode_closing_tag(INT8U *apdu, INT8U tag_number);

/* returns the true if the tag is an opening tag and the tag number matches */
BOOLEAN decode_is_opening_tag_number(INT8U *apdu, INT8U tag_number);
/* returns the true if the tag is a closing tag and the tag number matches */
BOOLEAN decode_is_closing_tag_number(INT8U *apdu, INT8U tag_number);


INT8U decode_tag_number(INT8U *apdu, INT8U *tag_number);
INT8U decode_tag_number_and_lvt(INT8U *apdu, INT8U *tag_number, INT32U *value);

/* returns true if the tag is context specific */
BOOLEAN decode_is_context_specific(INT8U *apdu);
/* returns true if the tag is context specific and matches */
BOOLEAN decode_is_context_tag(INT8U *apdu, INT8U tag_number);

/* from clause 20.2.14 Encoding of an Object Identifier Value */
/* and 20.2.1 General Rules for Encoding BACnet Tags */
/* returns the number of apdu bytes consumed */
INT16U encode_context_object_id(INT8U *apdu, INT8U tag_number, INT16U object_type, INT32U instance);
INT16U encode_tagged_object_id(INT8U *apdu, INT16U object_type, INT32U instance);

INT16U decode_context_object_id(INT8U *apdu, INT8U tag_number, INT16U *object_type, INT32U *instance);
INT16U decode_tagged_object_id(INT8U *apdu, INT16U *object_type, INT32U *instance);

/* from clause 20.2.9 Encoding of a Character String Value */
/* and 20.2.1 General Rules for Encoding BACnet Tags */
/* returns the number of apdu bytes consumed */
INT16U encode_context_char_string(INT8U *apdu, INT8U tag_number, INT8U *char_string);
INT16U encode_tagged_char_string(INT8U *apdu, INT8U *char_string);

INT16U decode_context_char_string(INT8U *apdu, INT8U tag_number, INT8U *char_string, INT16U str_len);
INT16U decode_tagged_char_string(INT8U *apdu, INT8U *char_string, INT16U str_len);
INT16U decode_context_char_str_ptr(INT8U *apdu, INT8U tag_number, INT8U **char_string, INT16U *str_len);
INT16U decode_tagged_char_str_ptr(INT8U *apdu, INT8U **char_string, INT16U *str_len);

/* from clause 20.2.10 Encoding of a Bit String Value */
/* returns the number of apdu bytes consumed */
INT16U encode_tagged_bit_string(INT8U *apdu, INT8U *bit_string, INT8U last_bit);
void   set_bit_in_string(INT8U *apdu, INT8U bit, BOOLEAN enabeled);


/* from clause 20.2.2 Encoding of Null */
/* returns the number of apdu bytes consumed */
INT16U encode_tagged_null(INT8U *apdu);

/* from clause 20.2.3 Encoding of a Boolean */
/* returns the number of apdu bytes consumed */
INT16U encode_tagged_boolean(INT8U *apdu, BOOLEAN value);
INT16U encode_context_boolean(INT8U *apdu, INT8U tag_number, BOOLEAN value);

INT16U decode_tagged_boolean(INT8U *apdu, BOOLEAN *value);

/* from clause 20.2.6 Encoding of a Real Number Value */
/* and 20.2.1 General Rules for Encoding BACnet Tags */
/* returns the number of apdu bytes consumed */
INT16U encode_tagged_real(INT8U *apdu, REAL value);
INT16U encode_context_real(INT8U *apdu, INT8U tag_number, REAL value);

INT16U decode_tagged_real(INT8U *apdu, REAL *real_value);

/* from clause 20.2.4 Encoding of an Unsigned Integer Value */
/* and 20.2.1 General Rules for Encoding BACnet Tags */
/* returns the number of apdu bytes consumed */
INT8U encode_context_unsigned(INT8U *apdu, INT8U tag_number, INT32U value);
INT8U encode_tagged_unsigned(INT8U *apdu, INT32U value);

INT8U decode_context_unsigned(INT8U *apdu, INT8U tag_number, INT32U *value);
INT8U decode_context_unsigned8(INT8U *apdu, INT8U tag_number, INT8U *value);
INT8U decode_tagged_unsigned(INT8U *apdu, INT32U *value);

/* from clause 20.2.5 Encoding of a Signed Integer Value */
/* and 20.2.1 General Rules for Encoding BACnet Tags */
/* returns the number of apdu bytes consumed */
INT16U encode_tagged_signed(INT8U *apdu, INT32S value);
INT16U encode_context_signed(INT8U *apdu, INT8U tag_number, INT32S value);

INT16U decode_context_signed(INT8U *apdu, INT8U tag_number, INT32S *value);
INT16U decode_tagged_signed(INT8U *apdu, INT32S *value);

/* from clause 20.2.11 Encoding of an Enumerated Value */
/* and 20.2.1 General Rules for Encoding BACnet Tags */
/* returns the number of apdu bytes consumed */
INT16U encode_tagged_enum(INT8U *apdu, INT32U value);
INT16U encode_context_enum(INT8U *apdu, INT8U tag_number, INT32U value);

INT16U decode_context_enum(INT8U *apdu, INT8U tag_number, INT32U *value);
INT16U decode_tagged_enum(INT8U *apdu, INT32U *value);

/* from clause 20.2.13 Encoding of a Time Value */
/* and 20.2.1 General Rules for Encoding BACnet Tags */
/* returns the number of apdu bytes consumed */
INT16U encode_tagged_time(INT8U *apdu, INT8U hour, INT8U min, INT8U sec, INT8U hundredths);
INT16U decode_bacnet_time(INT8U *apdu, INT8U *hour, INT8U *min, INT8U *sec, INT8U *hundredths);

/* BACnet Date */
/* year = years since 1900 */
/* month 1=Jan */
/* day = day of month */
/* wday 1=Monday...7=Sunday */

/* from clause 20.2.12 Encoding of a Date Value */
/* and 20.2.1 General Rules for Encoding BACnet Tags */
/* returns the number of apdu bytes consumed */
INT16U encode_tagged_date(INT8U *apdu, INT8U year, INT8U month, INT8U day, INT8U wday);
INT16U decode_date(INT8U *apdu, INT8U *year, INT8U *month, INT8U *day, INT8U *wday);

#endif
