/*####COPYRIGHTBEGIN####
 -------------------------------------------
 Copyright (C) 2005 Wapice Ltd

 Author: Kåre Särs (kare.sars@wapice.com)

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to:
 The Free Software Foundation, Inc.
 59 Temple Place - Suite 330
 Boston, MA  02111-1307
 USA.

 As a special exception, if other files instantiate templates or
 use macros or inline functions from this file, or you compile
 this file and link it with other works to produce a work based
 on this file, this file does not by itself cause the resulting
 work to be covered by the GNU General Public License. However
 the source code for this file must still be made available in
 accordance with section (3) of the GNU General Public License.

 This exception does not invalidate any other reasons why a work
 based on this file might be covered by the GNU General Public
 License.
 -------------------------------------------
####COPYRIGHTEND####*/

#include "platform.h"
#include "conf.h"
#include "npdu.h"
#include "apdu.h"
#include "bacenum.h"
#include "bacdcode.h"
#include "object.h"
#include "make_apdu.h"

/* from clause 20.1.4.3 in the BACnet stantard */
/**********************************************************************/
BACnet_npdu_t *apdu_simple_ack(BACnet_npdu_t *npdu, INT8U invoke_id, INT8U service)
{
    /* clear the old npdu data */
    init_npdu_data(npdu);

    npdu->data[0] = APDU_TYPE_SIMPLE_ACK;
    npdu->data[1] = invoke_id;
    npdu->data[2] = service;
    npdu->data_len = 3;

    return npdu;
}


/* from clause 20.1.7.4 in the BACnet stantard */
/**********************************************************************/
BACnet_npdu_t *apdu_error(BACnet_npdu_t *npdu, INT8U invoke_id, INT8U err_choice, INT8U class, INT8U code)
{
    /* clear the old npdu data */
    init_npdu_data(npdu);

    npdu->data[0] = APDU_TYPE_ERROR;
    npdu->data[1] = invoke_id;
    npdu->data[2] = err_choice;
    npdu->data_len = 3;

    npdu->data_len += encode_tagged_enum(&npdu->data[npdu->data_len], class);
    npdu->data_len += encode_tagged_enum(&npdu->data[npdu->data_len], code);

    return npdu;
}


/* from clause 20.1.9.4 in the BACnet stantard */
/**********************************************************************/
BACnet_npdu_t *apdu_abort(BACnet_npdu_t *npdu, INT8U invoke_id, INT8U reason)
{
    /* clear the old npdu data */
    init_npdu_data(npdu);

    npdu->data[0] = APDU_TYPE_ABORT | APDU_SRV;
    npdu->data[1] = invoke_id;
    npdu->data[2] = reason;
    npdu->data_len = 3;

    return npdu;
}

/* from clause 20.1.8.3 in the BACnet stantard */
/**********************************************************************/
BACnet_npdu_t *apdu_reject(BACnet_npdu_t *npdu, INT8U invoke_id, INT8U reason)
{
    /* clear the old npdu data */
    init_npdu_data(npdu);

    npdu->data[0] = APDU_TYPE_REJECT;
    npdu->data[1] = invoke_id;
    npdu->data[2] = reason;
    npdu->data_len = 3;

    return npdu;
}

/* I AM */
/* from clause 20.1.3.3 (Unconfirmed Request) */
/* and clause 16.10.3 and clause 21 Formal description... */
/**********************************************************************/
BACnet_npdu_t *apdu_iam(BACnet_npdu_t *npdu)
{
    /* clear the old npdu data */
    init_npdu_data(npdu);

    npdu->data[0] = APDU_TYPE_UNCONFIRMED_REQUEST;
    npdu->data[1] = SERVICE_UNCONFIRMED_I_AM;
    npdu->data_len = 2;

    /* add I-Am Device Identifier */
    npdu->data_len += encode_tagged_object_id(&(npdu->data[npdu->data_len]),
            OBJECT_DEVICE, ObjectList.node[0].instance);

    /* add Max APDU Length Accepted */
    npdu->data_len += encode_tagged_unsigned(&(npdu->data[npdu->data_len]), MAX_APDU_LEN);

    /* add Segmentation Supported */
    npdu->data_len += encode_tagged_enum(&(npdu->data[npdu->data_len]), SUPPORTED_SEGMENTATION);

    /* add Vendor Identifier */
    npdu->data_len += encode_tagged_unsigned(&(npdu->data[npdu->data_len]), ObjectList.node[0].type.dev->VendorId);

    /* "I am" is to be broadcasted localy or globaly */
    npdu->broadcast=GLOBAL_BROADCAST;

    return npdu;
}

/* I HAVE */
/* from clause 20.1.3.3 (Unconfirmed Request) */
/* and clause 16.9.3 and clause 21 Formal description... */
/**********************************************************************/
BACnet_npdu_t *apdu_ihave(BACnet_npdu_t *npdu, BACnet_object_t *object)
{
    /* did we get a valid object ? */
    if (object == NULL) return NULL;

    /* If we got here we have a real bacnet object */
    /* clear the old npdu data */
    init_npdu_data(npdu);

    npdu->data[0] = APDU_TYPE_UNCONFIRMED_REQUEST;
    npdu->data[1] = SERVICE_UNCONFIRMED_I_HAVE;
    npdu->data_len = 2;

    /* add Device Identifier */
    npdu->data_len += encode_tagged_object_id(&(npdu->data[npdu->data_len]),
            OBJECT_DEVICE, ObjectList.node[0].instance);

    /* add Object Identifier */
    npdu->data_len += encode_tagged_object_id(&(npdu->data[npdu->data_len]),
            object->obj_type, object->instance);

    /* add Object Name */
    npdu->data_len += encode_tagged_char_string(&(npdu->data[npdu->data_len]), object->name);

    /* "I have" is to be broadcasted localy or globaly as long as the requester gets it */
    npdu->broadcast=GLOBAL_BROADCAST;

    return npdu;
}


/* READ PROPERTY */
/* from clause 20.1.2.11 (Confirmed Request) */
/* and clause 15.5 and clause 21 Formal description... */
/**********************************************************************/
BACnet_npdu_t *apdu_read_property(BACnet_npdu_t *npdu, INT8U InvokeId,
                                  INT16U object_type, INT32U instance,
                                  BOOLEAN sa, INT8U max_segs, INT8U max_seg_size,
                                  INT8U property, INT8U *array)
{
    if (npdu==NULL) return NULL;

    /* clear the old npdu data */
    init_npdu_data(npdu);

    npdu->data[0] = APDU_TYPE_CONFIRMED_REQUEST;
    if (sa == TRUE) {
        npdu->data[0] |= APDU_SEG_ACC;
        if ((max_segs > 7) || (max_seg_size > 15)) return NULL;
        npdu->data[1] = (max_segs << 4) | max_seg_size;
    }
    else {
        npdu->data[1] = 0; /* no segmentation accepted */
    }
    npdu->data[2] = InvokeId;
    /* no segmentation => no seq_num or prop_win_size */
    npdu->data[3] = SERVICE_CONFIRMED_READ_PROPERTY;
    npdu->data_len=4;

    npdu->data_len += encode_context_object_id(&npdu->data[npdu->data_len], 0, object_type, instance);
    npdu->data_len += encode_context_unsigned(&npdu->data[npdu->data_len], 1, property);

    if (array != NULL) {
        npdu->data_len += encode_context_unsigned(&npdu->data[npdu->data_len], 2, *array);
    }

    return npdu;
}


/* READ PROPERTY ACK */
/* from clause 20.1.5.8 (Complex ACK) */
/* and clause 15.5 and clause 21 Formal description... */
/**********************************************************************/
BACnet_npdu_t *apdu_read_property_ack(BACnet_npdu_t *npdu, INT8U invoke_id, BACnet_object_t *object,
                                      INT32U property, INT32U array_index)
{
    INT16U len;

    /* did we get a valid object ? */
    if (object == NULL)
    {
        return apdu_error(npdu,
                          invoke_id,
                          SERVICE_CONFIRMED_READ_PROPERTY,
                          ERROR_CLASS_OBJECT,
                          ERROR_CODE_UNKNOWN_OBJECT);
    }

    /* clear the old npdu data */
    init_npdu_data(npdu);

    /* --- complex ACK header */
    npdu->data[0] = APDU_TYPE_COMPLEX_ACK;
    npdu->data[1] = invoke_id;
    npdu->data[2] = SERVICE_CONFIRMED_READ_PROPERTY;
    npdu->data_len = 3;

    /* --- Object Id */
    npdu->data_len += encode_context_object_id(&npdu->data[npdu->data_len], 0, object->obj_type, object->instance);
    /* --- Property */
    npdu->data_len += encode_context_enum(&npdu->data[npdu->data_len], 1, property);
    /* --- Array Index */
    if (array_index != BACNET_ARRAY_ALL) {
        npdu->data_len += encode_context_unsigned(&npdu->data[npdu->data_len], 2, array_index);
    }
    /* --- Value */
    /*   -- Opening tag */
    npdu->data_len += encode_opening_tag(&npdu->data[npdu->data_len], 3);
    /*   The Value */
    len = encode_object_value(&npdu->data[npdu->data_len], object, property, array_index);
    if (len == ENC_OBJ_ERR)
    {
        /* an error occured */
        return apdu_error(npdu,
                          invoke_id,
                          SERVICE_CONFIRMED_READ_PROPERTY,
                          ERROR_CLASS_PROPERTY,
                          npdu->data[npdu->data_len]);
    }
    /* If we get here all is OK */
    npdu->data_len += len;
    /*  -- Closing tag */
    npdu->data_len += encode_closing_tag(&npdu->data[npdu->data_len], 3);

    return npdu;
}

/* WHO IS */
/* from clause 20.1.3.3 (Unconfirmed Request) */
/* and clause 16.10.1 and clause 21 Formal description... */
/**********************************************************************/
BACnet_npdu_t *apdu_who_is(BACnet_npdu_t *npdu)
{
    if (npdu==NULL) return NULL;

    /* clear the old npdu data */
    init_npdu_data(npdu);

    npdu->data[0] = APDU_TYPE_UNCONFIRMED_REQUEST;
    npdu->data[1] = SERVICE_UNCONFIRMED_WHO_IS;
    npdu->data_len = 2;

    return npdu;
}

/* from clause 20.1.3.3 (Unconfirmed Request) */
/* and clause 16.10.1 and clause 21 Formal description... */
/**********************************************************************/
BACnet_npdu_t *apdu_who_is_ranged(BACnet_npdu_t *npdu, INT32U instance_low, INT32U instance_hi)
{
    if (npdu==NULL) return NULL;

    /* clear the old npdu data */
    init_npdu_data(npdu);

    npdu->data[0] = APDU_TYPE_UNCONFIRMED_REQUEST;
    npdu->data[1] = SERVICE_UNCONFIRMED_WHO_IS;
    npdu->data_len = 2;

    npdu->data_len += encode_context_unsigned(&npdu->data[npdu->data_len], 0, instance_low);
    npdu->data_len += encode_context_unsigned(&npdu->data[npdu->data_len], 1, instance_hi);

    return npdu;
}


/* WHO HAS */
/* from clause 20.1.3.3 (Unconfirmed Request) */
/* and clause 16.9.1 and clause 21 Formal description... */
/**********************************************************************/
BACnet_npdu_t *apdu_who_has(BACnet_npdu_t *npdu, INT16U object_type, INT32U instance, INT8U *name)
{
    if (npdu==NULL) return NULL;

    /* clear the old npdu data */
    init_npdu_data(npdu);

    npdu->data[0] = APDU_TYPE_UNCONFIRMED_REQUEST;
    npdu->data[1] = SERVICE_UNCONFIRMED_WHO_HAS;
    npdu->data_len = 2;


    if (name == NULL) {
        npdu->data_len += encode_context_object_id(&npdu->data[npdu->data_len], 2, object_type, instance);
    }
    else {
        npdu->data_len += encode_context_char_string(&npdu->data[npdu->data_len], 3, name);
    }

    return npdu;
}

/* from clause 20.1.3.3 (Unconfirmed Request) */
/* and clause 16.9.1 and clause 21 Formal description... */
/**********************************************************************/
BACnet_npdu_t *apdu_who_has_ranged(BACnet_npdu_t *npdu, INT32U instance_low, INT32U instance_hi,
                                 INT16U object_type, INT32U instance, INT8U *name)
{
    if (npdu==NULL) return NULL;

    /* clear the old npdu data */
    init_npdu_data(npdu);

    npdu->data[0] = APDU_TYPE_UNCONFIRMED_REQUEST;
    npdu->data[1] = SERVICE_UNCONFIRMED_WHO_HAS;
    npdu->data_len = 2;

    npdu->data_len += encode_context_unsigned(&npdu->data[npdu->data_len], 0, instance_low);
    npdu->data_len += encode_context_unsigned(&npdu->data[npdu->data_len], 1, instance_hi);

    if (name == NULL) {
        npdu->data_len += encode_context_object_id(&npdu->data[npdu->data_len], 2, object_type, instance);
    }
    else {
        npdu->data_len += encode_context_char_string(&npdu->data[npdu->data_len], 3, name);
    }

    return npdu;
}



/* from clause 20.1.2.11 (Confirmed Request) */
/* and clause 15.9 and clause 21 Formal description... */
/**********************************************************************/
BACnet_npdu_t *apdu_write_prop_real(BACnet_npdu_t *npdu, INT8U InvokeId,
                                    INT16U object_type, INT32U instance, INT8U property,
                                    REAL *value, INT8U *array, INT8U *priority)
{
    if (npdu==NULL) return NULL;

    /* clear the old npdu data */
    init_npdu_data(npdu);

    npdu->data[0] = APDU_TYPE_CONFIRMED_REQUEST;
    npdu->data[1] = 0; /* no segmentation */
    npdu->data[2] = InvokeId;
    /* no segmentation => no seq_num or prop_win_size */
    npdu->data[3] = SERVICE_CONFIRMED_WRITE_PROPERTY;
    npdu->data_len=4;

    /* object_id */
    npdu->data_len += encode_context_object_id(&npdu->data[npdu->data_len], 0, object_type, instance);
    /* property */
    npdu->data_len += encode_context_unsigned(&npdu->data[npdu->data_len], 1, property);
    /* array_index */
    if (array != NULL) {
        npdu->data_len += encode_context_unsigned(&npdu->data[npdu->data_len], 2, *array);
    }
    /* -- Opening tag */
    npdu->data_len += encode_opening_tag(&npdu->data[npdu->data_len], 3);
    /* value */
    if (value != NULL) npdu->data_len += encode_tagged_real(&npdu->data[npdu->data_len], *value);
    else               npdu->data_len += encode_tagged_null(&npdu->data[npdu->data_len]);
    /* -- Closing tag */
    npdu->data_len += encode_closing_tag(&npdu->data[npdu->data_len], 3);
    /* priority */
    if (priority != NULL) {
        npdu->data_len += encode_context_unsigned(&npdu->data[npdu->data_len], 4, *priority);
    }

    return npdu;
}



/* from clause 20.1.2.11 (Confirmed Request) */
/* and clause 15.9 and clause 21 Formal description... */
/**********************************************************************/
BACnet_npdu_t *apdu_write_prop_enum(BACnet_npdu_t *npdu, INT8U InvokeId,
                                    INT16U object_type, INT32U instance, INT8U property,
                                    INT8U *value, INT8U *array, INT8U *priority)
{
    if (npdu==NULL) return NULL;

    /* clear the old npdu data */
    init_npdu_data(npdu);

    npdu->data[0] = APDU_TYPE_CONFIRMED_REQUEST;
    npdu->data[1] = 0; /* no segmentation */
    npdu->data[2] = InvokeId;
    /* no segmentation => no seq_num or prop_win_size */
    npdu->data[3] = SERVICE_CONFIRMED_WRITE_PROPERTY;
    npdu->data_len=4;

    /* object_id */
    npdu->data_len += encode_context_object_id(&npdu->data[npdu->data_len], 0, object_type, instance);
    /* property */
    npdu->data_len += encode_context_unsigned(&npdu->data[npdu->data_len], 1, property);
    /* array_index */
    if (array != NULL) {
        npdu->data_len += encode_context_unsigned(&npdu->data[npdu->data_len], 2, *array);
    }
    /* -- Opening tag */
    npdu->data_len += encode_opening_tag(&npdu->data[npdu->data_len], 3);
    /* value */
    if (value != NULL) npdu->data_len += encode_tagged_enum(&npdu->data[npdu->data_len], *value);
    else               npdu->data_len += encode_tagged_null(&npdu->data[npdu->data_len]);
    /* -- Closing tag */
    npdu->data_len += encode_closing_tag(&npdu->data[npdu->data_len], 3);
    /* priority */
    if (priority != NULL) {
        npdu->data_len += encode_context_unsigned(&npdu->data[npdu->data_len], 4, *priority);
    }

    return npdu;
}



/* from clause 20.1.5 (Complex ACK PDU) */
/**********************************************************************/
BACnet_npdu_t *segmented_c_ack_pdu(BACnet_npdu_t *npdu, INT8U invoke_id, INT8U *data, INT16U data_len,
                                   INT16U segment_size, INT8U sequence_number, INT8U proposedWS)
{
    INT16U i, data_start, data_end, segment_data_size;

    if (npdu==NULL) return NULL;
    if (data==NULL) return NULL;
    /* clear the old npdu data */
    init_npdu_data(npdu);

    /* calculate starting point and end point */
    /* FIX_ME this has to be recalculated to correct sizes !!!!!!! */
    segment_data_size = segment_size - 5;
    data_start = ((sequence_number+0) * (segment_data_size)) + 3;
    data_end   = ((sequence_number+1) * (segment_data_size)) + 3;

    d_two_p_printf("data_start=%d, data_end=%d\n", data_start, data_end);
    if (data_end > data_len) data_end = data_len;
    d_two_p_printf("data_start=%d, data_end=%d\n", data_start, data_end);
    /* check if we have sent the last segment all ready */
    if (data_start >= data_len) return NULL;

    /* Header */
    npdu->data[0] = data[0] | APDU_SEGMENTED;
    if (data_end < data_len) npdu->data[0] |= APDU_MOR_SEGS;
    npdu->data[1] = invoke_id;
    npdu->data[2] = sequence_number;
    npdu->data[3] = proposedWS;
    npdu->data[4] = data[2];
    npdu->data_len = 5;

    /* Service Ack */
    for (i=0; i<(data_end - data_start); i++) {
        npdu->data[i+5] = data[i + data_start];
        npdu->data_len++;
    }

    /* now return this segment for sending */
    return npdu;
}


/* from clause 20.1.6 (Segment ACK PDU) */
/**********************************************************************/
BACnet_npdu_t *segment_ack_pdu(BACnet_npdu_t *npdu, INT8U invoke_id, INT8U sequence_number,
                               INT8U actualWS, BOOLEAN nak, BOOLEAN srv)
{
    if (npdu==NULL) return NULL;
    /* clear the old npdu data */
    init_npdu_data(npdu);

    npdu->data[0] = APDU_TYPE_SEGMENT_ACK;
    if (nak == TRUE) npdu->data[0] |= APDU_NAK;
    if (srv == TRUE) npdu->data[0] |= APDU_SRV;
    npdu->data[1] = invoke_id;
    npdu->data[2] = sequence_number;
    npdu->data[3] = actualWS;
    npdu->data_len = 4;

    /* now return this for sending */
    return npdu;
}

