/*####COPYRIGHTBEGIN####
 -------------------------------------------
 Copyright (C) 2004 Steve Karg

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to:
 The Free Software Foundation, Inc.
 59 Temple Place - Suite 330
 Boston, MA  02111-1307
 USA.

 As a special exception, if other files instantiate templates or
 use macros or inline functions from this file, or you compile
 this file and link it with other works to produce a work based
 on this file, this file does not by itself cause the resulting
 work to be covered by the GNU General Public License. However
 the source code for this file must still be made available in
 accordance with section (3) of the GNU General Public License.

 This exception does not invalidate any other reasons why a work
 based on this file might be covered by the GNU General Public
 License.

 Modified by Kåre Särs (kare.sars@wapice.com) March 2005
 -------------------------------------------
####COPYRIGHTEND####*/

#ifndef MSTP_H
#define MSTP_H

#define MAX_MSTP_PDU_SIZE 501

/*  The value 255 is used to denote broadcast when used as a */
/* destination address but is not allowed as a value for a station. */
#define MSTP_BROADCAST_ADDRESS 255

/* MS/TP Frame Type */
/* Frame Types 8 through 127 are reserved by ASHRAE. */
#define FRAME_TYPE_TOKEN 0
#define FRAME_TYPE_POLL_FOR_MASTER 1
#define FRAME_TYPE_REPLY_TO_POLL_FOR_MASTER 2
#define FRAME_TYPE_TEST_REQUEST 3
#define FRAME_TYPE_TEST_RESPONSE 4
#define FRAME_TYPE_BACNET_DATA_EXPECTING_REPLY 5
#define FRAME_TYPE_BACNET_DATA_NOT_EXPECTING_REPLY 6
#define FRAME_TYPE_REPLY_POSTPONED 7
/* Frame Types 128 through 255: Proprietary Frames */
/* These frames are available to vendors as proprietary (non-BACnet) frames. */
/* The first two octets of the Data field shall specify the unique vendor */
/* identification code, most significant octet first, for the type of */
/* vendor-proprietary frame to be conveyed. The length of the data portion */
/* of a Proprietary frame shall be in the range of 2 to 501 octets. */
#define FRAME_TYPE_PROPRIETARY_MIN 128
#define FRAME_TYPE_PROPRIETARY_MAX 255

/* This parameter represents the value of the Max_Info_Frames property of */
/* the node's Device object. The value of Max_Info_Frames specifies the */
/* maximum number of information frames the node may send before it must */
/* pass the token. Max_Info_Frames may have different values on different */
/* nodes. This may be used to allocate more or less of the available link */
/* bandwidth to particular nodes. If Max_Info_Frames is not writable in a */
/* node, its value shall be 1. */
#define MAX_INFO_FRAMES 1

/* This parameter represents the value of the Max_Master property of the */
/* node's Device object. The value of Max_Master specifies the highest */
/* allowable address for master nodes. The value of Max_Master shall be */
/* less than or equal to 127. If Max_Master is not writable in a node, */
/* its value shall be 127. */
#define MSTP_MAX_MASTER 127

/* receive FSM states */
typedef enum
{
    MSTP_RECEIVE_STATE_IDLE,
    MSTP_RECEIVE_STATE_PREAMBLE,
    MSTP_RECEIVE_STATE_HEADER,
    MSTP_RECEIVE_STATE_HEADER_CRC,
    MSTP_RECEIVE_STATE_DATA,
    MSTP_RECEIVE_STATE_DATA_CRC,
} MSTP_RECEIVE_STATE;

/* master node FSM states */
typedef enum
{
    MSTP_MASTER_STATE_INITIALIZE,           /* 0 */
    MSTP_MASTER_STATE_IDLE,                 /* 1 */
    MSTP_MASTER_STATE_USE_TOKEN,            /* 2 */
    MSTP_MASTER_STATE_WAIT_FOR_REPLY,       /* 3 */
    MSTP_MASTER_STATE_DONE_WITH_TOKEN,      /* 4 */
    MSTP_MASTER_STATE_PASS_TOKEN,           /* 5 */
    MSTP_MASTER_STATE_NO_TOKEN,             /* 6 */
    MSTP_MASTER_STATE_POLL_FOR_MASTER,      /* 7 */
    MSTP_MASTER_STATE_ANSWER_DATA_REQUEST,  /* 8 */
} MSTP_MASTER_STATE;

typedef enum
{
    MSTP_SEND_NOTHING,
    MSTP_SEND_NO_WAIT,
    MSTP_SEND_AND_WAIT,
    MSTP_SEND_REPLY,
    MSTP_SEND_TEST_REQUEST,
    MSTP_SEND_TEST_RESPONSE,
    MSTP_SEND_ERROR
} MSTP_WRITE_TYPE;


/* data for a given MS/TP port */
struct mstp_port_struct_t {
    /* Timer variables dependent on baudrate */
    INT8U Tframe_abort;
    INT8U Tturnaround;

    /* Do we have anything to send? */
    INT8U SendType;

    /* When this semafore is greater than one the read function can return */
    SEM_T *npdu_read_sem;
    /* This semafore ensures that only one instance of the read function is active at once */
    SEM_T *npdu_protect_read_sem;

    /* When this semafore is greater than one the write function can return */
    SEM_T *npdu_write_sem;
    /* This semafore ensures that only one instance of the write function is active at once */
    SEM_T *npdu_protect_write_sem;

    /* variables used for sending */
    /* ---------------------------------------------------------*/

    /* Are we in transmit mode or not */
    BOOLEAN TxEnabeled;

    /* MS/TP internal buffer for outgoing npdu frames. */
    INT8U O_Buffer[MSTP_PDU_SIZE];
    /* The size of the outgoing frame */
    INT16U O_Buffer_Len;
    /* The controll frame */
    INT8U O_Buffer_Controll[MSTP_HEADER_SIZE];

    /* Variables used in the actual sending of the frames (could be defined in rs485.*) */
    INT16U OutputLength;
    INT8U *OutputBuffer;

    /* variables used for passing the information to higher layers */
    /* the length of the received buffer */
    INT16U npdu_r_len;
    /* any errors that have accured during receive */
    INT16U npdu_r_err;
    /* The source address of the frame */
    INT8U  npdu_r_src_mac;
    /* Frame type (used for Needing reply / NOT needing reply) */
    INT8U  npdu_r_frame_type;
    /* Should we notify higher layers about ResetMaintenancePFM */
    /* (hack for sending I am when first connected to the network) */
    BOOLEAN notify_on_reset_pfm;

    INT16U looking_for_frame;
    INT16U found_frames;

    BOOLEAN run_msec_timer;

    /******** Variables used by the Frame State Machine ********/
    /*---------------------------------------------------------*/
    MSTP_RECEIVE_STATE receive_state;

    BOOLEAN DataAvailable; /* There is data in the buffer */

    BOOLEAN ReceiveError; /* TRUE when error detected during Rx octet */

    INT8U overrun;

    /* A Boolean flag set to TRUE by the Receive State Machine if an error is */
    /* detected during the reception of a frame. Set to FALSE by the main */
    /* state machine. */
    BOOLEAN ReceivedInvalidFrame;

    /* A Boolean flag set to TRUE by the Receive State Machine if a valid frame */
    /* is received. Set to FALSE by the main state machine. */
    BOOLEAN ReceivedValidFrame;

    /* After receiving a frame this value will be TRUE until Tturnaround has expired */
    /*BOOLEAN Turn_Around_Waiting;*/

    INT8U DataRegister; /* stores the latest data */

    INT16U DataCRC;     /* Used to accumulate the CRC on the data field of a frame. */

    INT16U DataLength;  /* Used to store the data length of a received frame. */

    INT8U DestinationAddress;  /* Used to store the destination address of a received frame. */

    INT8U FrameType; /* Used to store the frame type of a received frame. */

    /* Used to count the number of received octets or errors. */
    /* This is used in the detection of link activity. */
    INT16U EventCount;

    INT8U HeaderCRC;  /* Used to accumulate the CRC on the header of a frame. */

    /* Used as an index by the Send function, */
    /* up to a maximum value of MSTP_PDU_SIZE. */
    INT16U TxIndex;

    /* Used as an index by the Receive State Machine, */
    /* up to a maximum value of MSTP_PDU_SIZE. */
    INT16U RxIndex;

    /* A pointer to an array of octest where to save received frames. */
    /* This buffer is also used for sending test responce => MSTP_PDU_SIZE */
    INT8U  InputBuffer[MSTP_PDU_SIZE];

    /* A timer with nominal 5 millisecond resolution used to measure and */
    /* generate silence on the medium between octets. It is incremented by a */
    /* timer process and is cleared by the Receive State Machine when activity */
    /* is detected and by the SendFrame procedure as each octet is transmitted. */
    /* Since the timer resolution is limited and the timer is not necessarily */
    /* synchronized to other machine events, a timer value of N will actually */
    /* denote intervals between N-1 and N */
    INT16U SilenceTimer;

    INT8U SourceAddress; /* Used to store the Source Address of a received frame. */

    /* "This Station," the MAC address of this node. TS is generally read from a */
    /* hardware DIP switch, or from nonvolatile memory. Valid values for TS are */
    /* 0 to 254. The value 255 is used to denote broadcast when used as a */
    /* destination address but is not allowed as a value for TS. */
    /* Frame or Master??? */
    INT8U This_Station;

    /******** Variables used by the Master Node State Machine ********/
    /*---------------------------------------------------------------*/

    /* When a master node is powered up or reset, */
    /* it shall unconditionally enter the INITIALIZE state. */
    MSTP_MASTER_STATE master_state;

    /* The number of frames sent by this node during a single token hold. */
    /* When this counter reaches the value Nmax_info_frames, the node must */
    /* pass the token. */
    INT16U FrameCount;

    /* "Next Station," the MAC address of the node to which This Station passes */
    /* the token. If the Next_Station is unknown, Next_Station shall be equal to */
    /* This_Station. */
    INT8U Next_Station;

    /* "Poll Station," the MAC address of the node to which This Station last */
    /* sent a Poll For Master. This is used during token maintenance. */
    INT8U Poll_Station;

    /* A counter of transmission retries used for Token and Poll For Master */
    /* transmission. */
    INT16U RetryCount;

    /* A timer used to measure and generate Reply Postponed frames.  It is */
    /* incremented by a timer process and is cleared by the Master Node State */
    /* Machine when a Data Expecting Reply Answer activity is completed. */
    INT8U ReplyPostponedTimer;

    /* A Boolean flag set to TRUE by the master machine if this node is the */
    /* only known master node. */
    BOOLEAN SoleMaster;

    /* The number of tokens received by this node. When this counter reaches the */
    /* value Npoll, the node polls the address range between TS and NS for */
    /* additional master nodes. TokenCount is set to zero at the end of the */
    /* polling process. */
    INT16U TokenCount;

};


/* The minimum number of DataAvailable or ReceiveError events that must be */
/* seen by a receiving node in order to declare the line "active": 4. */
#define Nmin_octets 4

/* Read errors and messages returned in "type" */
/* (bits belonging to frame type 0xff) */
/* no PDU */
#define READ_ERR_LEN          0x0100
#define READ_ERR_REPL_TMO     0x0200
#define READ_TRY_NEXT_BR      0x0300
#define READ_RESTART_PFM      0x0400
/* can contain a pdu */
#define READ_ERR_MISSED_FRAME 0x8000

#define WRITE_ERR_LEN         0x0200
#define WRITE_ERR_WRONG_TYPE  0x0400
#define WRITE_ERR_WRONG_MAC   0x0800
#define WRITE_ERR_PREEMTIED   0x1000


/* initialization functions and tasks */
BOOLEAN MSTP_Init(struct mstp_port_struct_t *port, INT8U this_station_mac);
void MSTP_Re_Init(INT8U this_station_mac);
void MSTP_Millisecond_Timer(INT8U delay);
void MSTP_Receive_Frame_State_Machine(void);


/* Functions provided to the "User" */
INT16U MSTP_read(INT8U *buf, INT16U len, INT8U *from_mac, INT16U *type);

INT16U MSTP_write(INT8U *data, INT16U data_len, INT8U to_mac, MSTP_WRITE_TYPE type);


extern struct mstp_port_struct_t *mstp_port;

#endif
