/*####COPYRIGHTBEGIN####
 -------------------------------------------
 Copyright (C) 2005 Wapice Ltd

 Author: Kåre Särs (kare.sars@wapice.com)

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to:
 The Free Software Foundation, Inc.
 59 Temple Place - Suite 330
 Boston, MA  02111-1307
 USA.

 As a special exception, if other files instantiate templates or
 use macros or inline functions from this file, or you compile
 this file and link it with other works to produce a work based
 on this file, this file does not by itself cause the resulting
 work to be covered by the GNU General Public License. However
 the source code for this file must still be made available in
 accordance with section (3) of the GNU General Public License.

 This exception does not invalidate any other reasons why a work
 based on this file might be covered by the GNU General Public
 License.
 -------------------------------------------
####COPYRIGHTEND####*/

#include "platform.h"
#include "conf.h"
#include "bacenum.h"
#include "bacdcode.h"
#include "object.h"
#include "set_n_get.h"
#include "mstp.h"

/**********************************************************************/
BOOLEAN str_same(INT8U *str1, INT8U *str2)
{
    for (;;) {
        if ((*str1=='\0') && (*str2=='\0')) return TRUE;
        if (*str1!=*str2)                   return FALSE;
        if ((*str1=='\0') || (*str2=='\0')) return FALSE;
        str1++;
        str2++;
    }
    return FALSE;
}

/* return 0 on success and >0 on failure (the passed strings must be constant!)*/
/**********************************************************************/
INT8U init_analog_value(BACnet_analog_val_t *av, INT8U *name,
                       REAL def_val, INT8U units, INT8U cmd)
{
    BACnet_object_t *object;
    INT8U i;

    if (ObjectList.FirstBinaryValue != 0) {
        /* first analog objects then binary objects! */
        return 1;
    }
    if (ObjectList.FirstAnalogValue == 0) {
        ObjectList.FirstAnalogValue = ObjectList.NumDeviceObjects;
        ObjectList.NumAnalogValues=0;
    }

    object = &ObjectList.node[ObjectList.FirstAnalogValue + ObjectList.NumAnalogValues];

    /* fill the generic object values */
    object->obj_type =  OBJECT_ANALOG_VALUE;
    object->instance =  ObjectList.NumAnalogValues;
    object->name = name;
    object->type.ana_v = av;

    /* fill the analog value structure */
    object->type.ana_v->PresentValue = def_val;
    object->type.ana_v->StatusFlags[0] = 0; /* clear the bitstring */
    object->type.ana_v->EventState = EVENT_STATE_NORMAL;
    object->type.ana_v->OutOfService = FALSE;
    object->type.ana_v->Units = units;
    for (i=0; i<16; i++) object->type.ana_v->PriorityArray[i] = def_val;
    object->type.ana_v->PriorityFlags=0;
    object->type.ana_v->RelinquishDefault = def_val;
    object->type.ana_v->R_W_Commandable = cmd;

    ObjectList.NumAnalogValues++;
    ObjectList.Size++;

    return 0;
}


/* return 1 on success and 0 on failure (the passed strings must be constant!)*/
/**********************************************************************/
INT8U init_binary_value(BACnet_binary_val_t *bv, INT8U *name,
                        BOOLEAN def_val, INT8U *inactive, INT8U *active, INT8U cmd)
{
    BACnet_object_t *object;
    INT8U i;

    if (ObjectList.FirstBinaryValue == 0) {
        ObjectList.FirstBinaryValue = ObjectList.FirstAnalogValue + ObjectList.NumAnalogValues;
        ObjectList.NumBinaryValues=0;
    }

    object = &ObjectList.node[ObjectList.FirstBinaryValue + ObjectList.NumBinaryValues];

    /* fill the generic object values */
    object->obj_type =  OBJECT_BINARY_VALUE;
    object->instance =  ObjectList.NumBinaryValues;
    object->name = name;
    object->type.bin_v = bv;

    /* fill the binary value structure */
    object->type.bin_v->PresentValue = def_val;
    object->type.bin_v->StatusFlags[0] = 0; /* clear the bitstring */
    object->type.bin_v->EventState = EVENT_STATE_NORMAL;
    object->type.bin_v->OutOfService = FALSE;

    object->type.bin_v->ActiveText = active;
    object->type.bin_v->InactiveText = inactive;

    for (i=0; i<16; i++) object->type.bin_v->PriorityArray[i] = def_val;
    object->type.bin_v->PriorityFlags=0;
    object->type.bin_v->RelinquishDefault = def_val;

    object->type.bin_v->R_W_Commandable = cmd;

    ObjectList.NumBinaryValues++;
    ObjectList.Size++;

    return 0;
}

/* return the pointer to the referenced object or return NULL if not found */
/**********************************************************************/
BACnet_object_t *find_object_from_id(INT16U object_type, INT32U instance)
{

    switch (object_type)
    {
        case OBJECT_DEVICE:
            /* only one device and the first object */
            if ((instance != ObjectList.node[0].instance) &&
                 (instance != DEV_INST_HI_MAX))
            {
                return NULL;
            }
            return &ObjectList.node[0];

        case OBJECT_ANALOG_VALUE:
            /* is it a valid instance */
            if (instance >= ObjectList.NumAnalogValues) return NULL;

            return &ObjectList.node[ObjectList.FirstAnalogValue + instance];

        case OBJECT_BINARY_VALUE:
            /* is it a valid instance */
            if (instance >= ObjectList.NumBinaryValues) return NULL;

            return &ObjectList.node[ObjectList.FirstBinaryValue + instance];
    }

    return NULL;
}

/**********************************************************************/
BACnet_object_t *find_object_from_name(INT8U *object_name)
{
    INT8U i;

    for (i=0; i<ObjectList.Size; i++)
    {
        if (str_same(object_name, ObjectList.node[i].name) == TRUE)
        {
            return &ObjectList.node[i];
        }
    }

    return NULL;
}


/* On success this function returns the number of bytes used. */
/* On error this function returns ENC_OBJ_ERR and and an error code is */
/* returned in the first octet of the apdu. */
/**********************************************************************/
INT16U encode_object_value(INT8U *apdu, BACnet_object_t *object, INT32U property, INT32U index)
{
    REAL tmp_real;
    BOOLEAN tmp_bool;
    INT8U err;

    if (object == NULL) {
        apdu[0] = ERROR_CODE_UNKNOWN_OBJECT;
        return ENC_OBJ_ERR;
    }

    /* check array index */
    if (index != BACNET_ARRAY_ALL)
    {
        if ( ! ( (property == PROP_OBJECT_LIST) ||
                  (property == PROP_DEVICE_ADDRESS_BINDING) ||
                  (property == PROP_PRIORITY_ARRAY)) )
        {
            apdu[0] = ERROR_CODE_PROPERTY_IS_NOT_AN_ARRAY;
            return ENC_OBJ_ERR;
        }
    }

    switch (object->obj_type)
    {
        case OBJECT_DEVICE:
            switch (property)
            {
                case PROP_OBJECT_IDENTIFIER:
                    return encode_tagged_object_id(&apdu[0], OBJECT_DEVICE, ObjectList.node[0].instance);

                case PROP_OBJECT_NAME:
                    return encode_tagged_char_string(&apdu[0], object->name);

                case PROP_OBJECT_TYPE:
                    return encode_tagged_enum(&apdu[0], OBJECT_DEVICE);

                case PROP_SYSTEM_STATUS:
                    return encode_tagged_enum(&apdu[0], object->type.dev->SystemStatus);

                case PROP_VENDOR_NAME:
                    return encode_tagged_char_string(&apdu[0], object->type.dev->VendorName);

                case PROP_VENDOR_IDENTIFIER:
                    return encode_tagged_unsigned(&apdu[0], object->type.dev->VendorId);

                case PROP_MODEL_NAME:
                    return encode_tagged_char_string(&apdu[0], object->type.dev->ModelName);

                case PROP_FIRMWARE_REVISION:
                    return encode_tagged_char_string(&apdu[0], object->type.dev->FirmwareRev);

                case PROP_APPLICATION_SOFTWARE_VERSION:
                    return encode_tagged_char_string(&apdu[0], object->type.dev->ApplSoftwareRev);

                case PROP_PROTOCOL_VERSION:
                    return encode_tagged_unsigned(&apdu[0], BACNET_PROTOCOL_VERSION);

                case PROP_PROTOCOL_REVISION:
                    return encode_tagged_unsigned(&apdu[0], BACNET_PROTOCOL_REVISION);

                case PROP_PROTOCOL_SERVICES_SUPPORTED:
                    return encode_tagged_bit_string(&apdu[0],
                            object->type.dev->ServicesSupported,
                            BACNET_LAST_SERV_SUPP_BIT);

                case PROP_PROTOCOL_OBJECT_TYPES_SUPPORTED:
                    return encode_tagged_bit_string(&apdu[0],
                            object->type.dev->ObjectsSupported,
                            BACNET_LAST_OBJ_SUPP_BIT);

                case PROP_OBJECT_LIST:
                    if (index!=BACNET_ARRAY_ALL)
                    {
                        /* check index */
                        if (index > ObjectList.Size)
                        {
                            apdu[0]=ERROR_CODE_INVALID_ARRAY_INDEX;
                            return ENC_OBJ_ERR;
                        }
                        if (index==0) {
                            return encode_tagged_unsigned(&apdu[0], ObjectList.Size);
                        }
                        /* return the array element */
                        return  encode_tagged_object_id(&apdu[0],
                                ObjectList.node[index-1].obj_type,
                                ObjectList.node[index-1].instance);
                    }
                    else
                    {
                        INT16U i, len=0;

                        /* return the whole array */
                        for (i=0; i<ObjectList.Size; i++)
                        {
                            len+=encode_tagged_object_id(&apdu[len],
                                    ObjectList.node[i].obj_type,
                                    ObjectList.node[i].instance);
                        }
                        return len;
                    }

                case PROP_MAX_APDU_LENGTH_ACCEPTED:
                    return encode_tagged_unsigned(&apdu[0], MAX_APDU_LEN);

                case PROP_SEGMENTATION_SUPPORTED:
                    return encode_tagged_enum(&apdu[0], SUPPORTED_SEGMENTATION);

                case PROP_MAX_SEGMENTS_ACCEPTED:
                    return encode_tagged_unsigned(&apdu[0], MAX_SEGMENTS_ACCEPTED);

                case PROP_APDU_SEGMENT_TIMEOUT:
                    return encode_tagged_unsigned(&apdu[0], APDU_SEGMENT_TIMEOUT);

                case PROP_APDU_TIMEOUT:
                    return encode_tagged_unsigned(&apdu[0], 0);

                case PROP_NUMBER_OF_APDU_RETRIES:
                    return encode_tagged_unsigned(&apdu[0], 0);

                case PROP_MAX_MASTER:
                    return encode_tagged_unsigned(&apdu[0], MSTP_MAX_MASTER);

                case PROP_MAX_INFO_FRAMES:
                    return encode_tagged_unsigned(&apdu[0], MAX_INFO_FRAMES);

                case PROP_DEVICE_ADDRESS_BINDING:
                    return 0;

                case PROP_DATABASE_REVISION:
                    return encode_tagged_unsigned(&apdu[0], 0);

                default :
                    /* unsupported property */
                    apdu[0]=ERROR_CODE_UNKNOWN_PROPERTY;
                    return ENC_OBJ_ERR;
            }
            break;
        case OBJECT_ANALOG_VALUE:
            switch (property)
            {
                case PROP_OBJECT_IDENTIFIER:
                    return encode_tagged_object_id(&apdu[0], object->obj_type, object->instance);

                case PROP_OBJECT_NAME:
                    return encode_tagged_char_string(&apdu[0], object->name);

                case PROP_OBJECT_TYPE:
                    return encode_tagged_enum(&apdu[0], object->obj_type);

                case PROP_PRESENT_VALUE:
                    tmp_real = get_analog_present_value(object->instance, &err);
                    if (err==0) {
                        return encode_tagged_real(&apdu[0], tmp_real);
                    }
                    else {
                        apdu[0]=err;
                        return ENC_OBJ_ERR;
                    }

                case PROP_STATUS_FLAGS:
                    return encode_tagged_bit_string(&apdu[0],
                            object->type.ana_v->StatusFlags,
                            BACNET_LAST_STAT_FLAG_BIT);

                case PROP_EVENT_STATE:
                    return encode_tagged_enum(&apdu[0], object->type.ana_v->EventState);

                case PROP_OUT_OF_SERVICE:
                    return encode_tagged_boolean(&apdu[0], object->type.ana_v->OutOfService);

                case PROP_UNITS:
                    return encode_tagged_enum(&apdu[0], object->type.ana_v->Units);

                case PROP_PRIORITY_ARRAY:
                    if (object->type.ana_v->R_W_Commandable!=RWC_COMMANDABLE)
                    {
                        /* Not commandable ergo not readable */
                        apdu[0]=ERROR_CODE_UNKNOWN_PROPERTY;
                        return ENC_OBJ_ERR;
                    }
                    if (index!=BACNET_ARRAY_ALL)
                    {
                        /* check index */
                        if (index > 16)
                        {
                            apdu[0]=ERROR_CODE_INVALID_ARRAY_INDEX;
                            return ENC_OBJ_ERR;
                        }
                        else if (index == 0)
                        {
                            return encode_tagged_unsigned(&apdu[0], 16);
                        }
                        else if ((object->type.ana_v->PriorityFlags&(1<<(INT8U)(index-1)))==0)
                        {
                            return encode_tagged_null(&apdu[0]);
                        }
                        else
                        {
                            return encode_tagged_real(&apdu[0], object->type.ana_v->PriorityArray[index-1]);
                        }
                    }
                    else
                    {
                        INT8U i, len=0;

                        for (i=0; i<16; i++)
                        {
                            if ((object->type.ana_v->PriorityFlags&(1<<i))==0)
                            {
                                len+=encode_tagged_null(&apdu[len]);
                            }
                            else
                            {
                                len+=encode_tagged_real(&apdu[len], object->type.ana_v->PriorityArray[i]);
                            }
                        }
                        return len;
                    }

                case PROP_RELINQUISH_DEFAULT:
                    if (object->type.ana_v->R_W_Commandable!=RWC_COMMANDABLE)
                    {
                        /* Not commandable ergo not readable */
                        apdu[0]=ERROR_CODE_UNKNOWN_PROPERTY;
                        return ENC_OBJ_ERR;
                    }
                    return encode_tagged_real(&apdu[0], object->type.ana_v->RelinquishDefault);

                default :
                    /* unsupported property */
                    apdu[0]=ERROR_CODE_UNKNOWN_PROPERTY;
                    return ENC_OBJ_ERR;
            }
            break;
        case OBJECT_BINARY_VALUE:
            switch (property)
            {
                case PROP_OBJECT_IDENTIFIER:
                    return encode_tagged_object_id(&apdu[0], object->obj_type, object->instance);

                case PROP_OBJECT_NAME:
                    return encode_tagged_char_string(&apdu[0], object->name);

                case PROP_OBJECT_TYPE:
                    return encode_tagged_enum(&apdu[0], object->obj_type);

                case PROP_PRESENT_VALUE:
                    tmp_bool = get_binary_present_value(object->instance, &err);
                    if (err==0) {
                        return encode_tagged_enum(&apdu[0], tmp_bool);
                    }
                    else {
                        apdu[0] = err;
                        return ENC_OBJ_ERR;
                    }

                case PROP_STATUS_FLAGS:
                    return encode_tagged_bit_string(&apdu[0],
                            object->type.bin_v->StatusFlags,
                            BACNET_LAST_STAT_FLAG_BIT);

                case PROP_EVENT_STATE:
                    return encode_tagged_enum(&apdu[0], object->type.bin_v->EventState);

                case PROP_OUT_OF_SERVICE:
                    return encode_tagged_boolean(&apdu[0], object->type.bin_v->OutOfService);

                case PROP_INACTIVE_TEXT:
                    return encode_tagged_char_string(&apdu[0], object->type.bin_v->InactiveText);

                case PROP_ACTIVE_TEXT:
                    return encode_tagged_char_string(&apdu[0], object->type.bin_v->ActiveText);

                case PROP_PRIORITY_ARRAY:
                    if (object->type.bin_v->R_W_Commandable!=RWC_COMMANDABLE)
                    {
                        /* Not commandable ergo not readable */
                        apdu[0]=ERROR_CODE_UNKNOWN_PROPERTY;
                        return ENC_OBJ_ERR;
                    }
                    if (index!=BACNET_ARRAY_ALL)
                    {
                        /* check index */
                        if (index > 16)
                        {
                            apdu[0]=ERROR_CODE_INVALID_ARRAY_INDEX;
                            return ENC_OBJ_ERR;
                        }
                        else if (index == 0)
                        {
                            return encode_tagged_unsigned(&apdu[0], 16);
                        }
                        else if ((object->type.bin_v->PriorityFlags&(1<<(INT8U)(index-1)))==0)
                        {
                            return encode_tagged_null(&apdu[0]);
                        }
                        else
                        {
                            return encode_tagged_enum(&apdu[0], object->type.bin_v->PriorityArray[index-1]);
                        }
                    }
                    else
                    {
                        INT8U i, len=0;

                        for (i=0; i<16; i++)
                        {
                            if ((object->type.bin_v->PriorityFlags&(1<<i))==0)
                            {
                                len+=encode_tagged_null(&apdu[len]);
                            }
                            else
                            {
                                len+=encode_tagged_enum(&apdu[len], object->type.bin_v->PriorityArray[i]);
                            }
                        }
                        return len;
                    }

                case PROP_RELINQUISH_DEFAULT:
                    if (object->type.bin_v->R_W_Commandable!=RWC_COMMANDABLE)
                    {
                        /* Not commandable ergo not readable */
                        apdu[0]=ERROR_CODE_UNKNOWN_PROPERTY;
                        return ENC_OBJ_ERR;
                    }
                    return encode_tagged_enum(&apdu[0], object->type.bin_v->RelinquishDefault);

                default :
                    /* unsupported property */
                    apdu[0]=ERROR_CODE_UNKNOWN_PROPERTY;
                    return ENC_OBJ_ERR;
            }
            break;
    }

    /* if we get here we have forgot to implement something :) */
    apdu[0]=ERROR_CODE_UNKNOWN_OBJECT;
    return ENC_OBJ_ERR;
}

/**********************************************************************/
INT8U set_analog_object_value(BACnet_object_t *object, REAL *value, INT8U priority)
{
    INT8U i, ret, ret_old=0;
    priority--;

    // check the analog value range
    if ((value != NULL) && (check_analog_pv_range(object->instance, *value)!=0))
    {
        return ERROR_CODE_VALUE_OUT_OF_RANGE;
    }

    switch (object->type.ana_v->R_W_Commandable)
    {
        case RWC_COMMANDABLE:
            if (priority == 5) { /* 6-1  Priority level 6 is reserved clause 19.2.3 */
                return ERROR_CODE_WRITE_ACCESS_DENIED;
            }
            if (value == NULL) {
                object->type.ana_v->PriorityFlags &= ~(1<<priority);
            }
            else {
                object->type.ana_v->PriorityFlags |= (1<<priority);
                object->type.ana_v->PriorityArray[priority] = *value;
            }

            /* set the highest priorityArray[priority] value */
            for (i=0; i<16; i++) {
                if ((object->type.ana_v->PriorityFlags&(1<<i)) != 0) {
                    ret = set_analog_present_value(object->instance, object->type.ana_v->PriorityArray[i]);
                    if (ret != 0) {
                        /* this code will return the last error returned when trying to set the value */
                        object->type.ana_v->PriorityFlags &= ~(1<<i);
                        ret_old = ret;
                        continue;
                    }
                    return ret_old;
                }
            }
            /* if we get here set the relinguiched default (this MUST NOT fail !!) */
            (void)set_analog_present_value(object->instance, object->type.ana_v->RelinquishDefault);
            return ret_old;

        case RWC_READ_WRITE:
            if (value == NULL) {
                return ERROR_CODE_WRITE_ACCESS_DENIED;
            }
            return set_analog_present_value(object->instance, *value);

        case RWC_READ_ONLY:
            return ERROR_CODE_WRITE_ACCESS_DENIED;
    }
    return ERROR_CODE_WRITE_ACCESS_DENIED; /* fault */
}

/**********************************************************************/
INT8U set_binary_object_value(BACnet_object_t *object, BOOLEAN *value, INT8U priority)
{
    INT8U i, ret=0, ret_old=0;

    priority--;

    // we do not need to check the range here because it is already done in services.c

    switch (object->type.bin_v->R_W_Commandable)
    {
        case RWC_COMMANDABLE:
            if (priority == 5) { /* 6-1  Priority level 6 is reserved clause 19.2.3 */
                return ERROR_CODE_WRITE_ACCESS_DENIED;
            }
            if (value == NULL) {
                object->type.bin_v->PriorityFlags &= ~(1<<priority);
            }
            else {
                object->type.bin_v->PriorityFlags |= (1<<priority);
                object->type.bin_v->PriorityArray[priority] = *value;
            }
            /* set the highest priorityArray[priority] value */
            for (i=0; i<16; i++) {
                if ((object->type.bin_v->PriorityFlags&(1<<i)) != 0) {
                    ret = set_binary_present_value(object->instance, object->type.bin_v->PriorityArray[i]);
                    if (ret != 0) {
                        /* this code will return the last error returned when trying to set the value */
                        object->type.bin_v->PriorityFlags &= ~(1<<i);
                        ret_old = ret;
                        continue;
                    }
                    return ret_old;
                }
            }

            /* if we get here set the relinguiched default (this MUST NOT fail !!) */
            (void)set_binary_present_value(object->instance, object->type.bin_v->RelinquishDefault);
            return ret_old;

        case RWC_READ_WRITE:
            if (value == NULL) {
                return ERROR_CODE_WRITE_ACCESS_DENIED;
            }
            return set_binary_present_value(object->instance, *value);

        case RWC_READ_ONLY:
            return ERROR_CODE_WRITE_ACCESS_DENIED;
    }

    return ERROR_CODE_WRITE_ACCESS_DENIED; /* fault */
}
