/*####COPYRIGHTBEGIN####
 -------------------------------------------
 Copyright (C) 2005 Wapice Ltd

 Author: Kåre Särs (kare.sars@wapice.com)

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to:
 The Free Software Foundation, Inc.
 59 Temple Place - Suite 330
 Boston, MA  02111-1307
 USA.

 As a special exception, if other files instantiate templates or
 use macros or inline functions from this file, or you compile
 this file and link it with other works to produce a work based
 on this file, this file does not by itself cause the resulting
 work to be covered by the GNU General Public License. However
 the source code for this file must still be made available in
 accordance with section (3) of the GNU General Public License.

 This exception does not invalidate any other reasons why a work
 based on this file might be covered by the GNU General Public
 License.
 -------------------------------------------
####COPYRIGHTEND####*/

#ifndef BACNET_OBJECT_H
#define BACNET_OBJECT_H



#define BACNET_ARRAY_ALL 0xFFFFFF

#define RWC_READ_ONLY   0
#define RWC_READ_WRITE  1
#define RWC_COMMANDABLE 2

typedef struct {
    INT8U           SystemStatus;
    INT8U           *VendorName;
    INT16U          VendorId;
    INT8U           *ModelName;
    INT8U           *FirmwareRev;
    INT8U           *ApplSoftwareRev;
    INT8U           ServicesSupported[BACNET_SERV_SUPP_BSTR_LEN];
    INT8U           ObjectsSupported[BACNET_OBJ_SUPP_BSTR_LEN];
}BACnet_device_t;


typedef struct {
    REAL     PresentValue;
    INT8U    StatusFlags[1];
    INT8U    EventState;
    BOOLEAN  OutOfService;
    INT8U    R_W_Commandable;
    REAL     PriorityArray[16]; /* who commanded FIX_ME */
    INT16U   PriorityFlags; /* the flags start at 1<<0 (priority = [1-16]) */
    REAL     RelinquishDefault;
    INT8U    Units;
}BACnet_analog_val_t;


typedef struct {
    INT8U    PresentValue;
    INT8U    StatusFlags[1];
    INT8U    EventState;
    BOOLEAN  OutOfService;
    INT8U    R_W_Commandable;
    BOOLEAN  PriorityArray[16]; /* who commanded FIX_ME */
    INT16U   PriorityFlags; /* the flags start at 1<<0 (priority = [1-16]) */
    INT8U    RelinquishDefault;
    INT8U    *ActiveText;
    INT8U    *InactiveText;
}BACnet_binary_val_t;


typedef struct {

    INT16U obj_type;
    INT32U instance;
    INT8U  *name;

    union {
        BACnet_device_t *dev;
        BACnet_analog_val_t *ana_v;
        BACnet_binary_val_t *bin_v;
    } type;

} BACnet_object_t;


typedef struct {

    BACnet_object_t  *node;
    INT8U            Size;

    /* Device object */
    INT8U            NumDeviceObjects;

    /* Analog values */
    INT8U            FirstAnalogValue;
    INT8U            NumAnalogValues;

    /* Binary values */
    INT8U            FirstBinaryValue;
    INT8U            NumBinaryValues;

} BACnet_object_list;


extern BACnet_object_list ObjectList;


/* return 1 on success and 0 on failure (the passed strings must be constant!)*/
INT8U init_binary_value(BACnet_binary_val_t *bv, INT8U *name,
                        BOOLEAN def_val, INT8U *inactive, INT8U *active, INT8U cmd);

/* return 1 on success and 0 on failure (the passed strings must be constant!)*/
INT8U init_analog_value(BACnet_analog_val_t *av, INT8U *name,
                       REAL def_val, INT8U units, INT8U cmd);

BACnet_object_t *find_object_from_id(INT16U object_type, INT32U instance);
BACnet_object_t *find_object_from_name(INT8U *object_name);

/* On success this function returns the number of bytes used. */
/* On error this function returns 0 and and an error code is */
/* returned in the first octet of the apdu. */
INT16U encode_object_value(INT8U *apdu, BACnet_object_t *object, INT32U property, INT32U index);

INT8U set_analog_object_value(BACnet_object_t *object, REAL *value, INT8U priority);
INT8U set_binary_object_value(BACnet_object_t *object, BOOLEAN *value, INT8U priority);

#define ENC_OBJ_ERR 0x8000

#define DEV_INST_HI_MAX 4194303
#endif


