/*####COPYRIGHTBEGIN####
 -------------------------------------------
 Copyright (C) 2005 Wapice Ltd

 Author: Kåre Särs (kare.sars@wapice.com)

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to:
 The Free Software Foundation, Inc.
 59 Temple Place - Suite 330
 Boston, MA  02111-1307
 USA.

 As a special exception, if other files instantiate templates or
 use macros or inline functions from this file, or you compile
 this file and link it with other works to produce a work based
 on this file, this file does not by itself cause the resulting
 work to be covered by the GNU General Public License. However
 the source code for this file must still be made available in
 accordance with section (3) of the GNU General Public License.

 This exception does not invalidate any other reasons why a work
 based on this file might be covered by the GNU General Public
 License.
 -------------------------------------------
####COPYRIGHTEND####*/

#ifndef BAC_PLATFORM_H
#define BAC_PLATFORM_H

#undef IS_BIGENDIAN
#define SUPPORT_READ_PROPERTY_INIT
#define SUPPORT_ERROR_PDU_PRINT


/******** types ********/
typedef signed char         BOOLEAN;   /* 1 byte  0 or 1 */
#define TRUE 1
#define FALSE 0

typedef unsigned char       INT8U;     /* 1 byte  0 to 255 */
typedef char                INT8S;     /* 1 byte  -128 to 127 */
typedef unsigned short int  INT16U;    /* 2 bytes 0 to 65535 */
typedef unsigned int        INT32U;    /* 4 bytes 0 to 4294967295 */
typedef signed int          INT32S;    /* 4 bytes -2147483647 to 2147483647 +-1 :)*/

#define NULL            ((void *)0)

typedef float               REAL;     /* 4 bytes */


/******** Semafore handling ************/
#include <semaphore.h>  /* sem_t */
#include <stdlib.h>     /* malloc */

typedef sem_t                SEM_T;    /* semafore type */

#define bac_sem_wait(semafore) sem_wait(semafore)

#define bac_sem_post(semafore) sem_post(semafore)

#define bac_sem_trywait(semafore) sem_trywait(semafore)

#define bac_sem_init(semafore, value) {\
    semafore = malloc(sizeof(sem_t));\
    sem_init(semafore, 0, value);\
}

#define bac_sem_getvalue(semafore, semval) sem_getvalue(semafore, semval)


/* Here you could define functions to notify higher layers */
void indicate_sole_master(BOOLEAN soleMaster);
#define indicate_frame_error() { print_t(); printf("Frame error!!\n"); }
#define indicate_duplicated_mac() (void)printf("Somebody else is using the same MAC ID\n")
void try_next_baudrate(void);


/*********** printouts *****************/

#include <stdio.h> /* printf */
#include <sys/time.h>  /* gettimeofday */
#include <time.h>  /* time */

/* the d_ prefix is used for more verbose printouts */

/* Master node state maschine */
//#define ma_printf(args)
#ifndef ma_printf
#define ma_printf(args...) {struct timeval tv; gettimeofday(&tv,NULL); \
    printf("MAC %3d: M_SM: ", mstp_port->This_Station); \
    printf("%2d.%04d: ", (int)(tv.tv_sec&0xF),(int)tv.tv_usec/100); \
    printf("TMR: %3d: ", mstp_port->SilenceTimer); \
    printf("mstate: %d: ", mstp_port->master_state); \
    printf(args); }
#endif

/* Master node state maschine DEBUG */
#define d_ma_printf(args)
#ifndef d_ma_printf
#define d_ma_printf(args...) {struct timeval tv; gettimeofday(&tv,NULL); \
    printf("MAC %3d: MDSM: ", mstp_port->This_Station); \
    printf("%2d.%04d: ", (int)(tv.tv_sec&0xF),(int)tv.tv_usec/100); \
    printf("TMR: %3d: ", mstp_port->SilenceTimer); \
    printf("mstate: %d: ", mstp_port->master_state); \
    printf(args); }
#endif

/* FRAME STATE */
//#define fr_printf(args)
#ifndef fr_printf
#define fr_printf(args...) {struct timeval tv; gettimeofday(&tv,NULL); \
    printf("MAC %3d: F_SM: ", mstp_port->This_Station); \
    printf("%2d.%04d: ", (int)(tv.tv_sec&0xF),(int)tv.tv_usec/100); \
    printf("TMR: %3d: ", mstp_port->SilenceTimer); \
    printf("mstate: %d: ", mstp_port->master_state); \
    printf("Byte:0x%02x: ", mstp_port->DataRegister); \
    printf(args); }
#endif

/* FRAME STATE DEBUG*/
#define d_fr_printf(args...)
#ifndef d_fr_printf
#define d_fr_printf(args) {struct timeval tv; gettimeofday(&tv,NULL); \
    printf("MAC %3d: F_SM: ", mstp_port->This_Station); \
    printf("%2d.%04d: ", (int)(tv.tv_sec&0xF),(int)tv.tv_usec/100); \
    printf("Byte:0x%02x: ", mstp_port->DataRegister); \
    printf("TMR: %3d: ", mstp_port->SilenceTimer); \
    printf("mstate: %d: ", mstp_port->master_state); \
    printf(args); }
#endif


extern int pr;

#define cond_printf(args,...) {\
    if (pr==1) printf(args);\
    else {}\
}

#define print_t() { time_t t; struct tm tm; t=time(NULL); localtime_r(&t, &tm); \
    if (pr==1) printf("Time:%02d-%02d %2d:%02d:%02d: ", tm.tm_mon+1, tm.tm_mday, tm.tm_hour, tm.tm_min, tm.tm_sec);\
    else {}\
}

//#define NO_PRINTOUTS

#ifdef NO_PRINTOUTS

    #define no_p_printf(args)
    #define d_no_p_printf(args)

    #define one_p_printf(args, param)
    #define d_one_p_printf(args, param)

    #define two_p_printf(args, param1, param2)
    #define d_two_p_printf(args, param1, param2)

#else

    #define no_p_printf(args) cond_printf(args)
    #define d_no_p_printf(args)

    #define one_p_printf(args, param) {if (pr==1) {printf(args, param);}}
    #define d_one_p_printf(args, param)

    #define two_p_printf(args, param1, param2) {if (pr==1) {printf(args, param1, param2);}}
    #define d_two_p_printf(args, param1, param2)

#endif


#endif /* BAC_PLATFORM_H */
