/*####COPYRIGHTBEGIN####
 -------------------------------------------
 Copyright (C) 2005 Wapice Ltd

 Author: Kåre Särs (kare.sars@wapice.com)

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to:
 The Free Software Foundation, Inc.
 59 Temple Place - Suite 330
 Boston, MA  02111-1307
 USA.

 As a special exception, if other files instantiate templates or
 use macros or inline functions from this file, or you compile
 this file and link it with other works to produce a work based
 on this file, this file does not by itself cause the resulting
 work to be covered by the GNU General Public License. However
 the source code for this file must still be made available in
 accordance with section (3) of the GNU General Public License.

 This exception does not invalidate any other reasons why a work
 based on this file might be covered by the GNU General Public
 License.
 -------------------------------------------
####COPYRIGHTEND####*/


/* The module handles sending data out the RS-485 port */
/* and handles receiving data from the RS-485 port. */

#include <stdio.h>      /* printf */
#include <unistd.h>     /* read */
#include <sys/socket.h> /* socket */
#include <netinet/in.h> /* sockaddr_in */
#include <arpa/inet.h>  /* inet_pton, inet_ntoa */
#include <sys/time.h>   /* gettimeofday */
#include <fcntl.h> /* O_NDELAY... */
#include <termios.h>
#include <sys/ioctl.h>
//#include <pthread.h>

#include "platform.h"
#include "conf.h"
#include "npdu.h"

#include "mstp.h"
#include "rs485.h"

//#define DEBUG 1

static int sock;
INT8S rs485_device[20]="/dev/ttyS0";

SEM_T *rs485_send_sem;

/*************************************************************
* Returns the timedifference of two timestamps in usecs
*************************************************************/
int tv_diff(struct timeval tv1,struct timeval tv2) {
    long ds,dus;

    ds=(tv2.tv_sec&0x0000ffff)-(tv1.tv_sec&0x0000ffff);
    dus=(tv2.tv_usec)-(tv1.tv_usec);
    dus+=ds*1000000;

    return (int)dus;
}


/* Transmits a Frame on the wire */
/****************************************************************/
INT8U RS485_Send_First_Byte(INT8U *buffer, INT16U b_len)
{

    if (sock == 0) {
        printf("I'm not allowed to send\n");
        return 1; /* simulate interrupt stop */
    }
    mstp_port->OutputBuffer=buffer;
    mstp_port->OutputLength=b_len;

    mstp_port->TxIndex=0;

    /* we are now in send state */
    //printf("we are now in send state\n");
    mstp_port->TxEnabeled=TRUE;

    /* to simulate that we set SilenceTimer to 0 for each octet sent, */
    /* we stop the millisecond timer while sending */
    mstp_port->SilenceTimer=0;
    mstp_port->run_msec_timer=FALSE;

    bac_sem_post(rs485_send_sem);


    return 0;
}

/****************************************************************/
void *RS485_Send_Next_Byte(void *port)
{
    struct timeval tv_new, tv_old;

    bac_sem_init(rs485_send_sem, 0);

    for (;;)
    {
        bac_sem_wait(rs485_send_sem);

        while (mstp_port->TxIndex < mstp_port->OutputLength)
        {
            if (sock == 0) {
                printf("I'm not allowed to send\n");
                break; /* simulate interrupt stop */
            }
            write(sock, mstp_port->OutputBuffer + mstp_port->TxIndex,  1);
            #ifdef DEBUG
            printf("%02x ", mstp_port->OutputBuffer + mstp_port->TxIndex);
            #endif
            mstp_port->SilenceTimer=0;
            mstp_port->TxIndex++;
        }

        #ifdef DEBUG
        printf(" --> tcdrain time=%3d stimer=%d mstate=%d rstate=%d\n",
               (tv_diff(tv_old,tv_new) + 500)/1000,
               mstp_port->SilenceTimer,
               mstp_port->master_state,
               mstp_port->receive_state);
        #endif

        /* Wait until the final stop bit of the most significant CRC octet  */
        /* has been transmitted but not more than Tpostdrive. */
        gettimeofday(&tv_old, NULL);
        tcdrain(sock);
        gettimeofday(&tv_new, NULL);

        mstp_port->TxEnabeled=FALSE;

        mstp_port->run_msec_timer=TRUE;
    }
    return NULL;
}


/****************************************************************/
void *RS485_Interrupt_Handler(void *port) {
    INT8U tmp[1];
    INT32S ret;
    struct timeval tv_new, tv_old, tv;
    fd_set set;


    for(;;){

        tv.tv_sec=1;
        if (sock==0) {
            tv.tv_sec=0;
        }
        tv.tv_usec=0;
        FD_ZERO(&set);
        FD_SET(sock, &set);
        ret = select(FD_SETSIZE,&set,NULL,NULL,&tv);
        //printf("just returned from select with following value: %d\n", ret);

        /* simulate interrupt stop by closing the socket */
        if (sock==0) {
            //printf("I'm not allowed to read\n");
            ret=0;
        }

        if (ret < 0) {
            perror("select");
        }
        else if (ret>0) {

            ret=read(sock, tmp, 1);
            tv_old=tv_new;
            gettimeofday(&tv_new, NULL);

            if (ret!=1) {
                printf("r==%d sock=%d\n", ret, sock);
                perror("perror=");
            }

            if (mstp_port->TxEnabeled==TRUE){
                printf("dt=%8d received during Tx 0x%02x\n", tv_diff(tv_old,tv_new), tmp[0]);
            }
            else
            {
                /* copy data to variable */
                mstp_port->DataRegister = (INT8U)tmp[0];

                /* clear the silence timer */
                mstp_port->SilenceTimer = 0;

                /* data is ready, */
                mstp_port->DataAvailable = TRUE;

                /* process data */
                MSTP_Receive_Frame_State_Machine();

                /* In certain states the State Machine has to be run twice */
                if ((mstp_port->receive_state == MSTP_RECEIVE_STATE_HEADER_CRC) ||
                     (mstp_port->receive_state == MSTP_RECEIVE_STATE_DATA_CRC))
                {
                    MSTP_Receive_Frame_State_Machine();
                }
            }
        }
    }

    return NULL;
}




/****************************************************************/
int open_server(void) {
    struct sockaddr_in address;
    int sock=0;

    if ((sock = socket(AF_INET,SOCK_STREAM,0)) <= 0) {
        printf("Socket Creation FAILED\n");
        return 0;
    }
    address.sin_family = AF_INET;
    address.sin_port = htons(15000);
    inet_pton(AF_INET, "127.0.0.1" ,&address.sin_addr);

    if (connect(sock,(struct sockaddr *)&address,sizeof(address)) != 0) {
        printf("The connection was REJECTED from server %s...\n",inet_ntoa(address.sin_addr));
        close(sock);
    }

    return sock;
}


/****************************************************************/
int opentty(INT16U baudrate) {
    struct termios options;
    int tty_sock=0;

    /* open socket to ttyS0 */
    if ((tty_sock=open(rs485_device ,O_NDELAY | O_RDWR | O_NOCTTY))==-1) {
        printf("\"%s\" ",rs485_device);
        fflush(stdout);
        perror("");
        return 0;
    }

    /* Get the current options for the port... */
    tcgetattr(tty_sock, &options);

    /* Set (8N1) */
    options.c_iflag = IGNBRK;
    options.c_oflag = 0;
    options.c_lflag = 0;
    options.c_cflag = CS8 | CLOCAL | CREAD;

    /* Set the baud rates to 9600... */
    switch (baudrate)
    {
        case 38400:
            cfsetispeed(&options, B38400);
            cfsetospeed(&options, B38400);
            break;
        case 19200:
            cfsetispeed(&options, B19200);
            cfsetospeed(&options, B19200);
            break;
        case 9600:
            cfsetispeed(&options, B9600);
            cfsetospeed(&options, B9600);
            break;
        default:
            printf("wrong baudrate %d\n",baudrate);
            exit(-1);
    }

    /* Set the new options for the port... */
    tcsetattr(tty_sock, TCSANOW, &options);

    printf("tty_sock=%d\n", tty_sock);
    return tty_sock;
}


/****************************************************************/
void RS485_Stop_Uart(void) {
    //pthread_mutex_lock(&isr_mutex);
    printf("stopping uart\n");
    close(sock);
    sock = 0;
    //pthread_mutex_unlock(&isr_mutex);

}



/****************************************************************/
void RS485_Init_Uart(INT16U baudrate) {

    /* The minimum time without a DataAvailable or ReceiveError event within */
    /* a frame before a receiving node may discard the frame: 60 bit times. */
    /* (Implementations may use larger values for this timeout, */
    /* not to exceed 100 milliseconds.) */
    /* At 9600 baud, 60 bit times would be about 6.25 milliseconds */
    mstp_port->Tframe_abort=(INT8U)(1+ ((1000 * 60) / baudrate));
    printf("Tframe_abort=%d\n",mstp_port->Tframe_abort);

    /* The minimum time after the end of the stop bit of the final octet of a */
    /* received frame before a node may enable its EIA-485 driver: 40 bit times. */
    /* At 9600 baud, 40 bit times would be about 4.166 milliseconds */
    mstp_port->Tturnaround=(INT8U)(1+ ((1000 * 40) / baudrate));
    printf("Tturnaround=%d\n",mstp_port->Tturnaround);

    mstp_port->TxEnabeled=FALSE;

    sock = opentty(baudrate);
    //sock = open_server();

    if (sock<=0) {
        printf("EXITING!\n");
        exit(-1);
    }

    return;
}


