/*####COPYRIGHTBEGIN####
 -------------------------------------------
 Copyright (C) 2005 Wapice Ltd

 Author: Kåre Särs (kare.sars@wapice.com)

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to:
 The Free Software Foundation, Inc.
 59 Temple Place - Suite 330
 Boston, MA  02111-1307
 USA.

 As a special exception, if other files instantiate templates or
 use macros or inline functions from this file, or you compile
 this file and link it with other works to produce a work based
 on this file, this file does not by itself cause the resulting
 work to be covered by the GNU General Public License. However
 the source code for this file must still be made available in
 accordance with section (3) of the GNU General Public License.

 This exception does not invalidate any other reasons why a work
 based on this file might be covered by the GNU General Public
 License.
-------------------------------------------
####COPYRIGHTEND####*/

#include <stdio.h>
#include <sys/select.h>
#include<sys/socket.h>
#include<netinet/in.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h> /* O_NDELAY... */
#include <termios.h>
#include <sys/time.h>


#define USER_NOT_PRESENT 0
#define USER_PRESENT 1
#define MAX_CONNECTIONS 127

typedef struct {
    int socket;
    int status;
} User;

User users[MAX_CONNECTIONS];

/*************************************************************
* Returns the timedifference of two timestamps in usecs
*************************************************************/
int tv_diff(struct timeval tv1,struct timeval tv2) {
    long ds,dus;

    ds=(tv2.tv_sec&0x0000ffff)-(tv1.tv_sec&0x0000ffff);
    dus=(tv2.tv_usec)-(tv1.tv_usec);
    dus+=ds*1000000;

    return (int)dus;
}

/**************************************************************************/
int create_server_socket(int port) {
    int sock, reuse=1;

    struct sockaddr_in address;
    if ((sock = socket(AF_INET,SOCK_STREAM,0)) <= 0) printf("socket creation failed\n");
    setsockopt(sock, SOL_SOCKET, SO_REUSEADDR , &reuse , sizeof(reuse));

    address.sin_family = AF_INET;
    address.sin_addr.s_addr = INADDR_ANY;
    address.sin_port = htons(port);

    if (bind(sock,(struct sockaddr *)&address,sizeof(address)) != 0) printf("Binding Socket FAILED\n");

    return sock;
}

/**************************************************************************/
int main(void) {
    int server_socket, tty_sock;
    fd_set perm,set;
    struct sockaddr_in client_address;
    int ret, i, j;
    int nread;
    unsigned int client_len;
    int stdin_fno=0;
    char buffer[256];
    struct timeval tv, t_old, t_new;
    struct termios options;

    stdin_fno=fileno(stdin);
    printf("stdinfno=%d\n",stdin_fno);

    server_socket=create_server_socket(15000);
    if (server_socket<=0) return -1;

    listen(server_socket,10);

    /* open socket to ttyS0 */
    if ((tty_sock=open("/dev/ttyS14" ,O_NDELAY | O_RDWR | O_NOCTTY))==-1) {
        perror("opening /dev/ttyS14");
        return -1;
    }

    printf("tty_sock=%d\n", tty_sock);
    /* Get the current options for the port... */
    tcgetattr(tty_sock, &options);

    /* No parity (8N1) */
    options.c_iflag = IGNBRK;
    options.c_oflag = 0;
    options.c_lflag = 0;
    options.c_cflag = CS8 | CLOCAL | CREAD;

    /* Set the baud rates to 9600... */
    cfsetispeed(&options, B9600);
    cfsetospeed(&options, B9600);

    /* Set the new options for the port... */
    tcsetattr(tty_sock, TCSANOW, &options);


    FD_ZERO(&perm);
    FD_SET(server_socket, &perm);
    FD_SET(tty_sock, &perm);
    FD_SET(stdin_fno,&perm);

    while (1) {
        //printf("polling\n");
        tv.tv_sec=1;
        tv.tv_usec=0;
        set = perm;
        ret = select(FD_SETSIZE,&set,NULL,NULL,&tv);
        //printf("just returned from select with following value: %d\n", ret);
        if (ret < 0) {
            perror("select");
        }
        else  {
            //printf("select\n");
            if (FD_ISSET(server_socket, &set)) {
                printf("------- New user wants to connect -----------\n");
                /* find the first free entry */
                i = 0;
                for (i=0; i<MAX_CONNECTIONS; i++) {
                    if(users[i].status == USER_NOT_PRESENT) break;
                }
                if (i!=MAX_CONNECTIONS) {
                    users[i].socket = accept(server_socket, (struct sockaddr *)&client_address, &client_len);
                    FD_SET(users[i].socket, &perm);
                    users[i].status = USER_PRESENT;
                    printf("------- New user #%d connected on socket %d ------\n",i,users[i].socket);
                }
                else  printf("already MAX_CONNECTIONS users have applied!\n");
            }
            for (i=0; i<MAX_CONNECTIONS; i++) {
                if ((users[i].status!=USER_NOT_PRESENT) && (FD_ISSET(users[i].socket, &set))) {
                    ioctl(users[i].socket, FIONREAD, &nread);
                    if(nread == 0) {
                        printf("--------- user #%d closed socket %d ----------\n",i, users[i].socket);
                        close(users[i].socket);
                        FD_CLR(users[i].socket, &perm);
                        users[i].socket=0;
                        users[i].status=USER_NOT_PRESENT;
                    }
                    else {
                        /*printf("------- something came from the user on socket %d -------\n", users[i].socket);*/
                        do {
                            ret=read(users[i].socket, buffer, 255);
                            gettimeofday(&t_new, NULL);
                            /*buffer[ret]='\0';*/
                            for (j=0; j<MAX_CONNECTIONS; j++) {
                                if (users[j].status!=USER_NOT_PRESENT) {
                                    write(users[j].socket, buffer, ret);
                                }
                            }
                            printf("user[%d]=0x",i);
                            for (j=0; j<ret; j++) printf("%02x ",buffer[j]);
                            printf(" dt: %d\n", tv_diff(t_old, t_new));
                            t_old=t_new;

                            write(tty_sock, buffer, ret);
                            //printf("wrote to tty\n");

                            ioctl(users[i].socket, FIONREAD, &nread);

                        } while (nread != 0);
                    }
                }
            }
            if (FD_ISSET(stdin_fno,&set)) {
                printf("\n---------something came from stdin-----------\n");
                read(stdin_fno, buffer, 255);

                if ((strncmp(buffer,"quit",4)==0) || (strncmp(buffer,"exit",4)==0)) {
                    printf("Exiting!\n");
                    break;
                }
            }
            if (FD_ISSET(tty_sock, &set)) {
                ioctl(tty_sock, FIONREAD, &nread);
                if (nread == 0) {
                    printf("tty_sock closed!!\n");
                    return -1;
                }
                do {
                    ret=read(tty_sock, buffer, 255);
                    gettimeofday(&t_new, NULL);
                    printf("tty=0x");
                    for (j=0; j<ret; j++) printf("%02x ",buffer[j]);
                    printf(" dt: %d\n", tv_diff(t_old, t_new));
                    t_old=t_new;
                    for (j=0; j<MAX_CONNECTIONS; j++) {
                        if (users[j].status!=USER_NOT_PRESENT) {
                            write(users[j].socket, buffer, ret);
                            //printf("wrote to user[%d]\n", j);
                        }
                    }
                    ioctl(tty_sock, FIONREAD, &nread);
                } while (nread != 0);
            }
        }
    }
    return 0;
}

