/*####COPYRIGHTBEGIN####
 -------------------------------------------
 Copyright (C) 2005 Wapice Ltd

 Author: Kåre Särs (kare.sars@wapice.com)

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to:
 The Free Software Foundation, Inc.
 59 Temple Place - Suite 330
 Boston, MA  02111-1307
 USA.

 As a special exception, if other files instantiate templates or
 use macros or inline functions from this file, or you compile
 this file and link it with other works to produce a work based
 on this file, this file does not by itself cause the resulting
 work to be covered by the GNU General Public License. However
 the source code for this file must still be made available in
 accordance with section (3) of the GNU General Public License.

 This exception does not invalidate any other reasons why a work
 based on this file might be covered by the GNU General Public
 License.
 -------------------------------------------
####COPYRIGHTEND####*/

#include "platform.h"
#include "conf.h"
#include "npdu.h"
#include "apdu.h"
#include "bacenum.h"
#include "bacdcode.h"
#include "object.h"
#include "make_apdu.h"


/************************************************************************/
BACnet_npdu_t *handle_read_property(BACnet_npdu_t *npdu, INT8U InvokeId)
{
    INT16U object_type;
    INT32U instance;
    INT32U property;
    INT32U array_index=BACNET_ARRAY_ALL;
    INT8U *apdu;
    INT16U i=4; /* skip the confirmed service header (NO SEGMENTATION) */
    INT16U len;

    apdu=npdu->data;

    /* objectIdentifier */
    /* There should be at least one byte to read */
    if (i == npdu->data_len) {
        return apdu_reject(npdu, InvokeId, REJECT_REASON_MISSING_REQUIRED_PARAMETER);
    }
    i += len = decode_context_object_id(&apdu[i], 0, &object_type, &instance);
    if ((len==0) || (i > npdu->data_len)) {
        return apdu_reject(npdu, InvokeId, REJECT_REASON_INVALID_TAG);
    }

    /* propertyIdentifier */
    /* There should be at least one more byte to read */
    if (i == npdu->data_len) {
        return apdu_reject(npdu, InvokeId, REJECT_REASON_MISSING_REQUIRED_PARAMETER);
    }
    i += len = decode_context_enum(&apdu[i], 1, &property);
    if ((len==0) || (i > npdu->data_len)) {
        return apdu_reject(npdu, InvokeId, REJECT_REASON_INVALID_TAG);
    }

    /* propertyArrayIndex */
    if (npdu->data_len > i) {
        i += len = decode_context_unsigned(&apdu[i], 2, &array_index);
        if ((len==0) || (i > npdu->data_len)) {
            return apdu_reject(npdu, InvokeId, REJECT_REASON_INVALID_TAG);
        }
    }

    /* now if data_len still is > i, we have too many parameterts */
    if (npdu->data_len > i) {
        return apdu_reject(npdu, InvokeId, REJECT_REASON_TOO_MANY_ARGUMENTS);
    }

    /* check the object, retrive the value, create an apdu and return the apdu */
    return apdu_read_property_ack(npdu,
                                  InvokeId,
                                  find_object_from_id(object_type, instance),
                                  property,
                                  array_index);
}


/************************************************************************/
void handle_read_propertyACK(INT8U *apdu, INT16U apdu_len)
{
    INT16U  object_type;
    INT32U  instance;
    INT32U  property;
    INT32U  array_index=BACNET_ARRAY_ALL;
    INT16U  i=0;
    INT16U  len;
    REAL    ana_val=0;
    INT32U  bin_val=0;
    INT8U   string[80], tag;
    INT32S  sint;
    BOOLEAN bool;

    /* objectIdentifier */
    i += len = decode_context_object_id(&apdu[i], 0, &object_type, &instance);
    if (len==0) return;

    two_p_printf("object_type=%3d, instance=%3d, ",object_type, instance);

    /* propertyIdentifier */
    i += len = decode_context_enum(&apdu[i], 1, &property);
    if (len==0) return;

    one_p_printf("property=%3d ", property);

    (void)decode_tag_number(&apdu[i],  &tag);
    /* propertyArrayIndex */
    if (tag == 2) {
        len=decode_context_unsigned(&apdu[i], 2, &array_index);
        if (len==0) return;
    }
    d_one_p_printf("array_index=%3d ", array_index);

    /* opening tag */
    if (decode_is_opening_tag_number(&apdu[i], 3) == FALSE) {
        no_p_printf("Expected Opening tag 3\n");
    }
    i++; /* opening tag is one octet */

    /* data */
    if (decode_is_closing_tag_number(&apdu[i], 3) == TRUE) {
        /* check if we got an empty parameter */
        no_p_printf("No value passed\n\n");
        return;
    }

    while ((decode_is_closing_tag_number(&apdu[i], 3) == FALSE) && (i<apdu_len))
    {
        (void)decode_tag_number(&apdu[i], &tag);
        switch (tag)
        {
            case BACNET_APPLICATION_TAG_REAL:
                len = decode_tagged_real(&apdu[i], &ana_val);
                if (len==0) {no_p_printf("error reading tagged real value\n");}
                else  {one_p_printf("real value = %f\n", ana_val);}
                break;

            case BACNET_APPLICATION_TAG_ENUMERATED:
                len = decode_tagged_enum(&apdu[i], &bin_val);
                if (len==0) {no_p_printf("error reading tagged enumerated value\n");}
                else {one_p_printf("enumerated value = %d\n", bin_val);}
                break;

            case BACNET_APPLICATION_TAG_UNSIGNED_INT:
                len = decode_tagged_unsigned(&apdu[i], &bin_val);
                if (len==0){ no_p_printf("error reading tagged unsigned value\n");}
                else {one_p_printf("unsigned value = %d\n", bin_val);}
                break;

            case BACNET_APPLICATION_TAG_SIGNED_INT:
                len = decode_tagged_signed(&apdu[i], &sint);
                if (len==0) {no_p_printf("error reading tagged signed value\n");}
                else {one_p_printf("signed value = %d\n", sint);}
                break;

            case BACNET_APPLICATION_TAG_NULL:
                no_p_printf("NULL value\n");
                len = 1;
                break;

            case BACNET_APPLICATION_TAG_BOOLEAN:
                len = decode_tagged_boolean(&apdu[i], &bool);
                if (len==0) {no_p_printf("error reading tagged boolean value\n");}
                else {one_p_printf("boolean value = %d\n", bool);}
                break;

            case BACNET_APPLICATION_TAG_CHARACTER_STRING:
                len = decode_tagged_char_string(&apdu[i], &string[0], 80);
                if (len==0) {no_p_printf("error reading tagged string\n");}
                else {one_p_printf("string = \"%s\"\n", string);}
                break;

            case BACNET_APPLICATION_TAG_OBJECT_ID:
                len = decode_tagged_object_id(&apdu[i], &object_type, &instance);
                if (len==0) {no_p_printf("error reading ObjectId\n");}
                else {two_p_printf("ObjectType=%d Instance=%d\n", object_type, instance);}
                break;

            case BACNET_APPLICATION_TAG_BIT_STRING:
                len=0;
                no_p_printf("Bit String decoding and printing not supported yet!\n");
                break;

            default:
                one_p_printf("Unsupported Type %d\n", tag);
                len=0;
                /* Other types are not supported :) */
        }
        if (len == 0) break;
        else i += len;
    }
    if (i >= apdu_len) {
        two_p_printf("Unexpected end of apdu data, i=%d apdu_len=%d\n", i, apdu_len);
        return;
    }
    /* closing tag */
    if (decode_is_closing_tag_number(&apdu[i], 3) == FALSE)
    {
        no_p_printf("Expected Closing tag 3\n");
    }

    no_p_printf("\n");

    return;
}

/************************************************************************/
BACnet_npdu_t *handle_write_property(BACnet_npdu_t *npdu, INT8U InvokeId)
{
    INT16U  object_type;
    INT32U  instance;
    INT32U  property;
    INT32U  array_index=BACNET_ARRAY_ALL;
    INT8U  *apdu;
    INT16U  i=4; /* skip the confirmed service header (NO SEGMENTATION) */
    INT16U  len;
    INT8U   tag;
    REAL    ana_val=0;
    INT32U  bin_val=0;
    BOOLEAN bool_val;
    INT8U   priority=16;
    BACnet_object_t *object;
    INT8U err;

    apdu=npdu->data;

    no_p_printf("handle_write_property\n");
    /* objectIdentifier */
    /* There should be at least one byte to read */
    if (i == npdu->data_len) {
        return apdu_reject(npdu, InvokeId, REJECT_REASON_MISSING_REQUIRED_PARAMETER);
    }
    i += len = decode_context_object_id(&apdu[i], 0, &object_type, &instance);
    /* check the integrity of the read tag */
    if ((len==0) || (i > npdu->data_len)) {
        return apdu_reject(npdu, InvokeId, REJECT_REASON_INVALID_TAG);
    }

    two_p_printf("object_type=%d instance=%d ",(INT8U)object_type, (INT8U)instance);
    object = find_object_from_id(object_type, instance);
    if (object == NULL) {
        return apdu_error(npdu,
                          InvokeId,
                          SERVICE_CONFIRMED_WRITE_PROPERTY,
                          ERROR_CLASS_OBJECT,
                          ERROR_CODE_UNKNOWN_OBJECT);
    }

    /* there should be at least one more byte to read */
    if (i == npdu->data_len) {
        return apdu_reject(npdu, InvokeId, REJECT_REASON_MISSING_REQUIRED_PARAMETER);
    }
    /* propertyIdentifier */
    i += len = decode_context_enum(&apdu[i], 1, &property);
    if ((len==0) || (i > npdu->data_len)) {
        return apdu_reject(npdu, InvokeId, REJECT_REASON_INVALID_TAG);
    }

    //one_p_printf("property=%d\n", property);

    /* NOTE: We only support Present Value to be writabel. The switch case is for the error messages */
    if (property != PROP_PRESENT_VALUE)
    {
        switch (object_type)
        {
            case OBJECT_ANALOG_VALUE:
                switch (property)
                {
                    case PROP_OBJECT_IDENTIFIER:
                    case PROP_OBJECT_NAME:
                    case PROP_OBJECT_TYPE:
                    case PROP_STATUS_FLAGS:
                    case PROP_EVENT_STATE:
                    case PROP_OUT_OF_SERVICE:
                    case PROP_UNITS:
                    case PROP_PRIORITY_ARRAY:
                    case PROP_RELINQUISH_DEFAULT:
                        return apdu_error(npdu,
                                          InvokeId,
                                          SERVICE_CONFIRMED_WRITE_PROPERTY,
                                          ERROR_CLASS_PROPERTY,
                                          ERROR_CODE_WRITE_ACCESS_DENIED);
                    default :
                        /* unsupported property */
                        return apdu_error(npdu,
                                          InvokeId,
                                          SERVICE_CONFIRMED_WRITE_PROPERTY,
                                          ERROR_CLASS_PROPERTY,
                                          ERROR_CODE_UNKNOWN_PROPERTY);
                }
                break;

            case OBJECT_BINARY_VALUE:
                switch (property)
                {
                    case PROP_OBJECT_IDENTIFIER:
                    case PROP_OBJECT_NAME:
                    case PROP_OBJECT_TYPE:
                    case PROP_STATUS_FLAGS:
                    case PROP_EVENT_STATE:
                    case PROP_OUT_OF_SERVICE:
                    case PROP_INACTIVE_TEXT:
                    case PROP_ACTIVE_TEXT:
                    case PROP_PRIORITY_ARRAY:
                    case PROP_RELINQUISH_DEFAULT:
                        return apdu_error(npdu,
                                          InvokeId,
                                          SERVICE_CONFIRMED_WRITE_PROPERTY,
                                          ERROR_CLASS_PROPERTY,
                                          ERROR_CODE_WRITE_ACCESS_DENIED);
                    default :
                        /* unsupported property */
                        return apdu_error(npdu,
                                          InvokeId,
                                          SERVICE_CONFIRMED_WRITE_PROPERTY,
                                          ERROR_CLASS_PROPERTY,
                                          ERROR_CODE_UNKNOWN_PROPERTY);
                }
                break;

            case OBJECT_DEVICE:
                switch (property)
                {
                    case PROP_OBJECT_IDENTIFIER:
                    case PROP_OBJECT_NAME:
                    case PROP_OBJECT_TYPE:
                    case PROP_SYSTEM_STATUS:
                    case PROP_VENDOR_NAME:
                    case PROP_VENDOR_IDENTIFIER:
                    case PROP_MODEL_NAME:
                    case PROP_FIRMWARE_REVISION:
                    case PROP_APPLICATION_SOFTWARE_VERSION:
                    case PROP_PROTOCOL_VERSION:
                    case PROP_PROTOCOL_REVISION:
                    case PROP_PROTOCOL_SERVICES_SUPPORTED:
                    case PROP_PROTOCOL_OBJECT_TYPES_SUPPORTED:
                    case PROP_OBJECT_LIST:
                    case PROP_MAX_APDU_LENGTH_ACCEPTED:
                    case PROP_SEGMENTATION_SUPPORTED:
                    case PROP_MAX_SEGMENTS_ACCEPTED:
                    case PROP_APDU_SEGMENT_TIMEOUT:
                    case PROP_APDU_TIMEOUT:
                    case PROP_NUMBER_OF_APDU_RETRIES:
                    case PROP_MAX_MASTER:
                    case PROP_MAX_INFO_FRAMES:
                    case PROP_DEVICE_ADDRESS_BINDING:
                    case PROP_DATABASE_REVISION:
                        return apdu_error(npdu,
                                          InvokeId,
                                          SERVICE_CONFIRMED_WRITE_PROPERTY,
                                          ERROR_CLASS_PROPERTY,
                                          ERROR_CODE_WRITE_ACCESS_DENIED);
                    default :
                        /* unsupported property */
                        return apdu_error(npdu,
                                          InvokeId,
                                          SERVICE_CONFIRMED_WRITE_PROPERTY,
                                          ERROR_CLASS_PROPERTY,
                                          ERROR_CODE_UNKNOWN_PROPERTY);
                }
                break;

            default:
                return apdu_error(npdu,
                                  InvokeId,
                                  SERVICE_CONFIRMED_WRITE_PROPERTY,
                                  ERROR_CLASS_OBJECT,
                                  ERROR_CODE_UNKNOWN_OBJECT);
        }
    } /* if (property != PROP_PRESENT_VALUE) */

    /* propertyArrayIndex */
    /* there should be at least one more byte to read */
    if (i == npdu->data_len) {
        return apdu_reject(npdu, InvokeId, REJECT_REASON_MISSING_REQUIRED_PARAMETER);
    }
    /* check if we have an Array Index */
    (void)decode_tag_number(&apdu[i],  &tag);
    if (tag == 2) {
        i += len = decode_context_unsigned(&apdu[i], 2, &array_index);
        if ((len==0) || (i > npdu->data_len)) {
            return apdu_reject(npdu, InvokeId, REJECT_REASON_INVALID_TAG);
        }

        /* there should be at least one more byte to read */
        if (i == npdu->data_len) {
            return apdu_reject(npdu, InvokeId, REJECT_REASON_MISSING_REQUIRED_PARAMETER);
        }
    }
    one_p_printf("array_index=%d\n", array_index);

    /* NOTE: Writing to an array is not supported */
    if (array_index != BACNET_ARRAY_ALL) {
        return apdu_error(npdu,
                          InvokeId,
                          SERVICE_CONFIRMED_WRITE_PROPERTY,
                          ERROR_CLASS_PROPERTY,
                          ERROR_CODE_PROPERTY_IS_NOT_AN_ARRAY);
    }

    /* Parse Value */
    /* opening tag */
    if (decode_is_opening_tag_number(&apdu[i], 3) == FALSE) {
        /* Expected opening tag */
        return  apdu_reject(npdu, InvokeId, REJECT_REASON_INVALID_TAG);
    }
    i++; /* one octet opening tag */

    /* data */
    /* there should be at least one more byte to read */
    if (i == npdu->data_len) {
        return apdu_reject(npdu, InvokeId, REJECT_REASON_MISSING_REQUIRED_PARAMETER);
    }
    (void)decode_tag_number(&apdu[i],  &tag);
    switch (tag)
    {
        case BACNET_APPLICATION_TAG_REAL:
            i += len = decode_tagged_real(&apdu[i], &ana_val);
            if ((len==0) || (i > npdu->data_len)) {
                return apdu_reject(npdu, InvokeId, REJECT_REASON_INVALID_TAG);
            }
            break;
        case BACNET_APPLICATION_TAG_ENUMERATED:
            i += len = decode_tagged_enum(&apdu[i], &bin_val);
            if ((len==0) || (i > npdu->data_len)) {
                return apdu_reject(npdu, InvokeId, REJECT_REASON_INVALID_TAG);
            }
            break;
        case BACNET_APPLICATION_TAG_NULL:
            i++; /* null tag is one octet */
            break;
        default:
            /* Other types are not supported :) */
            return apdu_error(npdu,
                              InvokeId,
                              SERVICE_CONFIRMED_WRITE_PROPERTY,
                              ERROR_CLASS_PROPERTY,
                              ERROR_CODE_INVALID_DATA_TYPE);
    }

    /* closing tag */
    /* there should be at least one more byte to read */
    if (i == npdu->data_len) {
        return apdu_reject(npdu, InvokeId, REJECT_REASON_MISSING_REQUIRED_PARAMETER);
    }
    if (decode_is_closing_tag_number(&apdu[i], 3) == FALSE)
    {
        /* Expected closing tag */
        return  apdu_reject(npdu, InvokeId, REJECT_REASON_INVALID_TAG);
    }
    i++; /* only one octet for closing tag */

    /* Priority */
    if (npdu->data_len > i) {
        i += len = decode_context_unsigned8(&apdu[i], 4, &priority);
        if ((len==0) || (i > npdu->data_len)) {
            return apdu_reject(npdu, InvokeId, REJECT_REASON_INVALID_TAG);
        }
    }
    if ((priority==0) || (priority>16)) {
        return apdu_reject(npdu, InvokeId, REJECT_REASON_PARAMETER_OUT_OF_RANGE);
    }
    one_p_printf("priority=%d\n", priority);

    /* Done parsing the message, try writing to the object */


    switch (object_type)
    {
        case OBJECT_ANALOG_VALUE:
            /* NOTE: Present Value is writable if Out_Of_Service is true */
            if ((object->type.ana_v->R_W_Commandable == RWC_READ_ONLY) &&
                 (object->type.ana_v->OutOfService == FALSE) ) {
                return apdu_error(npdu,
                                  InvokeId,
                                  SERVICE_CONFIRMED_WRITE_PROPERTY,
                                  ERROR_CLASS_PROPERTY,
                                  ERROR_CODE_WRITE_ACCESS_DENIED);
            }

            /* set the value */
            if (tag == BACNET_APPLICATION_TAG_NULL) {
                err=set_analog_object_value(object, NULL, priority);
            }
            else if (tag == BACNET_APPLICATION_TAG_REAL) {
                err=set_analog_object_value(object, &ana_val, priority);
            }
            else {
                return apdu_error(npdu,
                                  InvokeId,
                                  SERVICE_CONFIRMED_WRITE_PROPERTY,
                                  ERROR_CLASS_PROPERTY,
                                  ERROR_CODE_INVALID_DATA_TYPE);
            }
            /* Did the writing fail? */
            if (err != 0) {
                return apdu_error(npdu,
                                  InvokeId,
                                  SERVICE_CONFIRMED_WRITE_PROPERTY,
                                  ERROR_CLASS_PROPERTY,
                                  err);
            }
            else {
                return apdu_simple_ack(npdu, InvokeId, SERVICE_CONFIRMED_WRITE_PROPERTY);
            }

        case OBJECT_BINARY_VALUE:
            if ((object->type.bin_v->R_W_Commandable == RWC_READ_ONLY) &&
                 (object->type.ana_v->OutOfService == FALSE) ) {
                return apdu_error(npdu,
                                  InvokeId,
                                  SERVICE_CONFIRMED_WRITE_PROPERTY,
                                  ERROR_CLASS_PROPERTY,
                                  ERROR_CODE_WRITE_ACCESS_DENIED);
            }

            if (tag == BACNET_APPLICATION_TAG_NULL) {
                err = set_binary_object_value(object, NULL, priority);
            }
            else if (tag == BACNET_APPLICATION_TAG_ENUMERATED) {
                if (bin_val > 1) {
                    return apdu_error(npdu,
                                      InvokeId,
                                      SERVICE_CONFIRMED_WRITE_PROPERTY,
                                      ERROR_CLASS_PROPERTY,
                                      ERROR_CODE_VALUE_OUT_OF_RANGE);
                }
                bool_val = (bin_val == 0) ? FALSE : TRUE;
                err = set_binary_object_value(object, &bool_val, priority);
            }
            else {
                return apdu_error(npdu,
                                  InvokeId,
                                  SERVICE_CONFIRMED_WRITE_PROPERTY,
                                  ERROR_CLASS_PROPERTY,
                                  ERROR_CODE_INVALID_DATA_TYPE);
            }

            /* Did the writing fail? */
            if (err != 0) {
                return apdu_error(npdu,
                                  InvokeId,
                                  SERVICE_CONFIRMED_WRITE_PROPERTY,
                                  ERROR_CLASS_PROPERTY,
                                  err);
            }
            else {
                return apdu_simple_ack(npdu, InvokeId, SERVICE_CONFIRMED_WRITE_PROPERTY);
            }

            /* check the object, retrive the value, create an apdu and return the apdu */
        default :
            return apdu_error(npdu,
                              InvokeId,
                              SERVICE_CONFIRMED_WRITE_PROPERTY,
                              ERROR_CLASS_OBJECT,
                              ERROR_CODE_UNKNOWN_OBJECT);
    }

    return npdu;
}


/************************************************************************/
/* Check if we should return I_Am and return a pointer to a npdu if yes */
/************************************************************************/
BACnet_npdu_t *handle_who_is(BACnet_npdu_t *npdu)
{
    INT32U DevInstLow=0;
    INT32U DevInstHi=DEV_INST_HI_MAX;

    /* Check for a range (header == 2 bytes) */
    if (npdu->data_len > 2)
    {
        INT8U *apdu;
        INT8U i=2, len;
        apdu=npdu->data;

        /* context-tagged DevInstLow */
        i += len = decode_context_unsigned(&apdu[i], 0, &DevInstLow);
        if (len==0) return NULL; /* something is wrong */

        /* context-tagged DevInstHi */
        i += len = decode_context_unsigned(&apdu[i], 1, &DevInstHi);
        if (len==0) return NULL; /* something is wrong */


    }

    if ((DevInstLow<=ObjectList.node[0].instance) && (DevInstHi>=ObjectList.node[0].instance))
    {
        return apdu_iam(npdu);
    }

    return NULL;
}


/************************************************************************/
BACnet_npdu_t *handle_who_has(BACnet_npdu_t *npdu)
{
    INT32U DevInstLow=0;
    INT32U DevInstHi=DEV_INST_HI_MAX;
    INT8U *apdu;
    INT16U i=2;
    INT16U len=0;
    INT8U  tag=0;
    INT16U object_type;
    INT32U instance;
    INT8U  *char_string;
    INT16U str_len=0;

    apdu=npdu->data;

    /* check if we have a range specified */
    (void)decode_tag_number(&apdu[i],  &tag);
    if (tag==0) {
        /* we have a range specified */

        /* context-tagged DevInstLow */
        i += len = decode_context_unsigned(&apdu[i], 0, &DevInstLow);
        if (len==0) return NULL; /* something is wrong */

        /* context-tagged DevInstHi */
        i += len = decode_context_unsigned(&apdu[i], 1, &DevInstHi);
        if (len==0) return NULL; /* something is wrong */

        if ((DevInstLow > ObjectList.node[0].instance) || (DevInstHi < ObjectList.node[0].instance)) {
            /* we are not in the range */
            return NULL;
        }
    }

    /* check if we have objectIdentifier or objectName */
    (void)decode_tag_number(&apdu[i],  &tag);
    switch (tag) {
        case 2:
            /* objectIdentifier */
            len=decode_context_object_id(&apdu[i], 2, &object_type, &instance);
            if (len==0) return NULL; /* something is wrong */

            return apdu_ihave(npdu, find_object_from_id(object_type, instance));
            break;
        case 3:
            /* objectName */
            len = decode_context_char_str_ptr(&apdu[i], 3, &char_string, &str_len);
            if (len==0) return NULL; /* something is wrong */
            /* decode_context_char_str_ptr does not add string termination and returns
               a pointer to the apdu data. The apdu is not needed any longer so we
               add string termination to the end of the string now */
            char_string[str_len]='\0';

            return apdu_ihave(npdu, find_object_from_name(char_string));
            break;
    }

    return NULL;
}


/************************************************************************/
void handle_i_have(BACnet_npdu_t *npdu)
{
    INT16U object_type=0;
    INT32U instance=0;
    INT8U  char_string[60];
    INT16U len=0;
    INT16U i=2;
    INT8U *apdu;

    apdu=npdu->data;

    /* skip the apdu header */

    /* device id */
    i += len=decode_tagged_object_id(&apdu[i], &object_type, &instance);
    if (len == 0) return;
    two_p_printf("device_type=%d instance=%d\n", object_type, instance);

    /* object id */
    i += len=decode_tagged_object_id(&apdu[i], &object_type, &instance);
    if (len == 0) return;
    two_p_printf("object_type=%d instance=%d\n", object_type, instance);

    len=decode_tagged_char_string(&apdu[i], char_string, 60);
    if (len == 0) return;
    one_p_printf("Name=%s\n", char_string);

    return;
}

/************************************************************************/
void handle_i_am(BACnet_npdu_t *npdu)
{
    INT16U object_type=0;
    INT32U instance=0;
    INT16U len=0;
    INT16U i=2;
    INT8U *apdu;

    print_t();
    no_p_printf("Received I'am\n");

    apdu=npdu->data;

    /* skip the apdu header */

    /* device id */
    i += len=decode_tagged_object_id(&apdu[i], &object_type, &instance);
    if (len == 0) return;
    two_p_printf("device_type=%d instance=%d\n", object_type, instance);

    /* max APDU Length */
    i += len=decode_tagged_unsigned(&apdu[i], &instance);
    if (len == 0) return;
    one_p_printf("Max Apdu Length = %d\n", instance);

    /* segmentation */
    i += len=decode_tagged_enum(&apdu[i], &instance);
    if (len == 0) return;
    one_p_printf("Segmentation = %d\n", instance);

    /* vendorId */
    i += len=decode_tagged_unsigned(&apdu[i], &instance);
    if (len == 0) return;
    one_p_printf("Vendor ID = %d\n", instance);

    return;
}

